import os
import datetime
import numpy as np
import matplotlib.pyplot as plt
from numpy.polynomial import Polynomial

from scipy.integrate import solve_ivp
from scipy.integrate._ivp.ivp import METHODS as ODE_METHODS


from . import graphics
from . import utilities as utils
from . import pysolver_view as psv
from . import fluid_properties as props


COLORS_MATLAB = graphics.COLORS_MATLAB
PROCESS_TYPES = ["polytropic", "adiabatic"]
PROPERTY_CALCULATION_TYPES = ["blending", "equilibrium", "metastable"]

from functools import wraps


class BarotropicModelOneComponent:
    """
    Coordinates the simulation and polynomial fitting for a barotropic process.

    Parameters
    ----------
    fluid_name : str
        The name of the fluid to be used in the simulation (e.g., 'Water').
    T_in : float
        Inlet temperature of the fluid in Kelvin.
    p_in : float
        Inlet pressure of the fluid in Pascals.
    p_out : float
        Outlet pressure of the fluid in Pascals.
    efficiency : float
        The efficiency of the polytropic process, dimensionless.
    calculation_type : str, optional
        The type of calculation to perform. Options include:

        - ``equilibrium``: Computes equilibrium properties only.
        - ``metastable``: Computes metastable properties only.
        - ``blending``: Computes both equilibrium and metastable properties, blends them, and returns the blended properties.

        Default is None, which will raise an error.
    blending_onset : float, optional
        The onset of blending in the process, typically a value between 0 and 1. Required when `calculation_type` is ``blending``.
    blending_width : float, optional
        The width of the blending region, typically a value between 0 and 1. Required when `calculation_type` is ``blending``.
    ODE_solver : str, optional
        The solver to use for the ODE integration. Valid options:

        .. list-table::
            :widths: 20 50
            :header-rows: 1

            * - Solver name
              - Description
            * - ``RK23``
              - Explicit Runge-Kutta method of order 3(2)
            * - ``RK45``
              - Explicit Runge-Kutta method of order 5(4)
            * - ``DOP853``
              - Explicit Runge-Kutta method of order 8
            * - ``Radau``
              - Implicit Runge-Kutta method of the Radau IIA family of order 5
            * - ``BDF``
              - Implicit multi-step variable-order (1 to 5) method based on a backward differentiation formula for the derivative approximation
            * - ``LSODA``
              - Adams/BDF method with automatic stiffness detection and switching

        See `Scipy solver_ivp() <https://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.solve_ivp.html>`_  for more info.
        Default is ``LSODA``.
        
        Recommended solvers: ``BDF``, ``LSODA``, or ``Radau`` for stiff problems or ``RK45`` for non-stiff problems with smooth blending.

    ODE_tolerance : float, optional
        The relative and absolute tolerance for the ODE solver. Default is 1e-8.
    HEOS_solver : str, optional
        The solver algorithm used to compute the metastable states. Valid options:

        .. list-table::
            :widths: 20 50
            :header-rows: 1

            * - Solver name
              - Description
            * - ``hybr``
              -  Powell's hybrid trust-region algorithm
            * - ``lm``
              - Levenberg–Marquardt algorithm

        See `Scipy root() <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.root.html>`_ for more info. Default is ``hybr``.
        
        Recommended solvers: Both ``hybr`` and ``lm`` work well for the tested cases.
    HEOS_tolerance : float, optional
        The tolerance for the HEOS solver. Default is 1e-6.
    HEOS_max_iter : int, optional
        The maximum number of iterations for the HEOS solver. Default is 100.
    HEOS_print_convergence : bool, optional
        If True, prints convergence information for the HEOS solver. Default is False.
    polynomial_degree : int
        Degree of the polynomials to fit. Default is 8.

        .. note::

            When `calculation_type` is ``blending`` the degree of the polynomial in the blending region is set to 4 to achieve sufficient accuracy while preventing numerical round-off errors associated with single-precision arithmetic in CFD solvers.

    polynomial_format : str, optional
        Type of polynomial representation (``horner`` or ``standard``). Default is 'horner'.
    polynomial_variables : list of str
        A list of variable names to fit polynomials to, such as 'density', 'viscosity',
        'speed_sound', 'void_fraction', 'vapor_quality'.
    output_dir : str
        The directory where output will be saved.

    """

    def __init__(
        self,
        fluid_name: str,
        T_in: float,
        p_in: float,
        p_out: float,
        efficiency: float = 1.00,
        calculation_type: str = None,
        blending_onset: float = None,
        blending_width: float = None,
        ODE_solver: str = "lsoda",
        ODE_tolerance: float = 1e-8,
        HEOS_solver: str = "hybr",
        HEOS_tolerance: float = 1e-6,
        HEOS_max_iter: int = 100,
        HEOS_print_convergence: bool = False,
        polynomial_degree: int = 8,
        polynomial_format: str = "horner",
        polynomial_variables: list = [
            "density",
            "viscosity",
            "speed_sound",
            # "void_fraction",
            # "vapor_quality",
        ],
        output_dir: str = "barotropic_model",
    ):

        # Rename mandatory arguments
        self.fluid_name = fluid_name
        self.T_in = T_in
        self.p_in = p_in
        self.p_out = p_out

        # Assign optional arguments with defaults
        self.efficiency = efficiency
        self.calculation_type = calculation_type
        self.blending_onset = blending_onset
        self.blending_width = blending_width
        self.ODE_solver = ODE_solver
        self.ODE_tolerance = ODE_tolerance
        self.HEOS_solver = HEOS_solver
        self.HEOS_tolerance = HEOS_tolerance
        self.HEOS_max_iter = HEOS_max_iter
        self.HEOS_print_convergence = HEOS_print_convergence
        self.polynomial_degree = polynomial_degree
        self.polynomial_format = polynomial_format
        self.polynomial_variables = polynomial_variables
        self.output_dir = output_dir

        # Initialize variables
        self.states = None
        self.ode_solution = None
        self.poly_fitter = None

    def compute_properties(self):
        """
        Solves the ODE for the barotropic process and stores the states.

        See Also
        --------
        barotropic_model_one_fluid : 
            Function used to compute the fluid properties.
        """
        # Manually pass all parameters to the ODE solver function
        self.states, self.ode_solution = barotropic_model_one_fluid(
            fluid_name=self.fluid_name,
            T_in=self.T_in,
            p_in=self.p_in,
            p_out=self.p_out,
            efficiency=self.efficiency,
            calculation_type=self.calculation_type,
            blending_onset=self.blending_onset,
            blending_width=self.blending_width,
            ODE_solver=self.ODE_solver,
            ODE_tolerance=self.ODE_tolerance,
            HEOS_solver=self.HEOS_solver,
            HEOS_tolerance=self.HEOS_tolerance,
            HEOS_max_iter=self.HEOS_max_iter,
            HEOS_print_convergence=self.HEOS_print_convergence,
        )

    def fit_polynomials(self):
        """
        Fits polynomials to the states using the PolynomialFitter class.
        
        See Also
        --------
        PolynomialFitter : 
            Class used for generating fitting polynomials.
        """

        if not self.states and not self.ode_solution:
            msg = "Fluid properties are not computed. Please call the method 'compute_properties()' first."
            raise ValueError(msg)

        # Fit the polynomials
        self.poly_fitter = PolynomialFitter(
            states=self.states,
            variables=self.polynomial_variables,
            degree=self.polynomial_degree,
            calculation_type=self.calculation_type,
            output_dir=self.output_dir,
        )
        self.poly_fitter.fit_polynomials()

        # Automatically initialize the ExpressionExporter
        self.exporter = ExpressionExporter(
            poly_fitter=self.poly_fitter,
            poly_format=self.polynomial_format,
        )

    def export_expressions_fluent(self, output_dir=None):
        """
        Exports the polynomial expressions in a format suitable for Ansys Fluent.

        Parameters
        ----------
        output_dir : str, optional
            The directory where the expressions will be saved. It uses a default directory if not provided.

        See Also
        --------
        ExpressionExporter : 
            Class for exporting polynomial expressions for use in CFD software.
        """
        if not self.exporter:
            msg = "ExpressionExporter not initialized. Please call the method 'fit_polynomials()' first."
            raise ValueError(msg)

        output_dir = output_dir or self.output_dir
        self.exporter.export_expressions_fluent(output_dir=output_dir)

    def export_expressions_cfx(self, output_dir=None):
        """
        Exports the polynomial expressions in a format suitable for Ansys CFX.

        Parameters
        ----------
        output_dir : str, optional
            The directory where the expressions will be saved. It uses a default directory if not provided.

        See Also
        --------
        ExpressionExporter : 
            Class for exporting polynomial expressions for use in CFD software.
        """
        if not self.exporter:
            msg = "ExpressionExporter not initialized. Please call the method 'fit_polynomials()' first."
            raise ValueError(msg)

        output_dir = output_dir or self.output_dir
        self.exporter.export_expressions_cfx(output_dir=output_dir)


def barotropic_model_one_fluid(
    fluid_name,
    T_in,
    p_in,
    p_out,
    efficiency,
    calculation_type=None,
    blending_onset=None,
    blending_width=None,
    ODE_solver="lsoda",
    ODE_tolerance=1e-8,
    HEOS_solver="hybr",
    HEOS_tolerance=1e-6,
    HEOS_max_iter=100,
    HEOS_print_convergence=False,
):
    r"""
    Simulates a polytropic process for a given fluid from an inlet state to a specified outlet pressure.

    This function integrates the differential equation describing the polytropic process, which may involve phase changes, using the relationship:

    .. math::

        \frac{\text{d}h}{\text{d}p} = \frac{\eta_p}{\rho}

    where :math:`h` is the specific enthalpy, :math:`p` is the pressure, :math:`\eta_p` is the process efficiency, and :math:`\rho` is the density of the fluid.

    The evaluation of fluid properties varies based on the ``calculation_type`` parameter:

    - ``calculation_type='equilibrium'``: Computes properties assuming thermodynamic equilibrium using the CoolProp property solver.
    - ``calculation_type='metastable'``: Computes properties based on the Helmholtz energy equations of state, iterating on density and temperature.
    - ``calculation_type='blending'``: Blends equilibrium and metastable properties according to the fluid vapor quality.

    In single-phase regions, all calculation types yield identical results. The differences arise in the evaluation within the two-phase region, where the handling of equilibrium and metastable fluid properties differs.

    .. note::

        The choice of ODE solver can significantly impact the accuracy and stability of the simulation. The ``BDF`` and ``LSODA`` solvers are recommended for general use, particularly for stiff problems involving equilibrium phase change or narrow blending widths, while `RK45` may be suitable for non-stiff problems with no phase change or smooth phase transitions.

    Parameters
    ----------
    fluid_name : str
        The name of the fluid to be used in the simulation (e.g., 'Water').
    T_in : float
        Inlet temperature of the fluid in Kelvin.
    p_in : float
        Inlet pressure of the fluid in Pascals.
    p_out : float
        Outlet pressure of the fluid in Pascals.
    efficiency : float
        The efficiency of the polytropic process, dimensionless.
    calculation_type : str, optional
        The type of calculation to perform. Options include:

        - ``equilibrium``: Computes equilibrium properties only.
        - ``metastable``: Computes metastable properties only.
        - ``blending``: Computes both equilibrium and metastable properties, blends them, and returns the blended properties.

        Default is None, which will raise an error.
    blending_onset : float, optional
        The onset of blending in the process, typically a value between 0 and 1. Required when `calculation_type` is ``blending``.
    blending_width : float, optional
        The width of the blending region, typically a value between 0 and 1. Required when `calculation_type` is ``blending``.
    ODE_solver : str, optional
        The solver to use for the ODE integration. Valid options:

        .. list-table::
            :widths: 20 50
            :header-rows: 1

            * - Solver name
              - Description
            * - ``RK23``
              - Explicit Runge-Kutta method of order 3(2)
            * - ``RK45``
              - Explicit Runge-Kutta method of order 5(4)
            * - ``DOP853``
              - Explicit Runge-Kutta method of order 8
            * - ``Radau``
              - Implicit Runge-Kutta method of the Radau IIA family of order 5
            * - ``BDF``
              - Implicit multi-step variable-order (1 to 5) method based on a backward differentiation formula for the derivative approximation
            * - ``LSODA``
              - Adams/BDF method with automatic stiffness detection and switching

        See `Scipy solver_ivp() <https://docs.scipy.org/doc/scipy/reference/generated/scipy.integrate.solve_ivp.html>`_  for more info.
        Default is ``LSODA``.
        
        Recommended solvers: ``BDF``, ``LSODA``, or ``Radau`` for stiff problems or ``RK45`` for non-stiff problems with smooth blending.

    ODE_tolerance : float, optional
        The relative and absolute tolerance for the ODE solver. Default is 1e-8.
    HEOS_solver : str, optional
        The solver algorithm used to compute the metastable states. Valid options:

        .. list-table::
            :widths: 20 50
            :header-rows: 1

            * - Solver name
              - Description
            * - ``hybr``
              -  Powell's hybrid trust-region algorithm
            * - ``lm``
              - Levenberg–Marquardt algorithm

        See `Scipy root() <https://docs.scipy.org/doc/scipy/reference/generated/scipy.optimize.root.html>`_ for more info. Default is ``hybr``.
        
        Recommended solvers: Both ``hybr`` and ``lm`` work well for the tested cases.
    HEOS_tolerance : float, optional
        The tolerance for the HEOS solver. Default is 1e-6.
    HEOS_max_iter : int, optional
        The maximum number of iterations for the HEOS solver. Default is 100.
    HEOS_print_convergence : bool, optional
        If True, prints convergence information for the HEOS solver. Default is False.

    Returns
    -------
    states : list
        A list of states (dictionaries) representing the properties of the fluid at each evaluation point.
    solution : scipy.integrate.OdeResult
        The result of the ODE integration containing information about the solver process.

    Examples
    --------
    >>> states, solution = barotropic_model_one_fluid(
    ...     fluid_name='Water',
    ...     T_in=300.0,
    ...     p_in=101325.0,
    ...     p_out=200000.0,
    ...     efficiency=0.85,
    ...     calculation_type='blending',
    ...     blending_onset=0.1,
    ...     blending_width=0.05,
    ...     ODE_solver='LSODA',
    ...     ODE_tolerance=1e-8,
    ...     HEOS_solver='hybr',
    ...     HEOS_tolerance=1e-6,
    ...     HEOS_max_iter=100,
    ...     HEOS_print_convergence=False,
    ...     n_eval=100
    ... )
    """

    # Initialize fluid and compute inlet state
    fluid = props.Fluid(name=fluid_name, backend="HEOS", exceptions=True)
    state_in = fluid.get_state(
        props.PT_INPUTS, p_in, T_in, supersaturation=True, generalize_quality=True
    )

    # Determine the type of phase change process
    phase_change_type = (
        "flashing" if state_in.s < fluid.critical_point.s else "condensation"
    )

    # Initial guess for the metastable state (could be updated later)
    rhoT_guess_metastable = [state_in.rho, state_in.T]

    # Differential equation defining the polytropic process
    def odefun(t, y):
        nonlocal rhoT_guess_metastable  # Allows modification within the function scope

        # Rename input variables
        p = t
        (h,) = y

        # Do calculations according to specified type
        if calculation_type == "equilibrium":
            # Compute equilibrium state using CoolProp solver
            state_eq = fluid.get_state(
                input_type=props.HmassP_INPUTS,
                prop_1=h,
                prop_2=p,
                generalize_quality=True,
                supersaturation=True,
            )
            dhdp = np.atleast_1d(efficiency / state_eq["rho"])
            return dhdp, state_eq

        elif calculation_type == "metastable":
            # Compute metastable state using custom solver
            state_meta = fluid.get_state_metastable(
                prop_1="h",
                prop_1_value=h,
                prop_2="p",
                prop_2_value=p,
                rhoT_guess=rhoT_guess_metastable,
                generalize_quality=False,
                supersaturation=True,
                solver_algorithm=HEOS_solver,
                solver_tolerance=HEOS_tolerance,
                solver_max_iterations=HEOS_max_iter,
                print_convergence=HEOS_print_convergence,
            )

            # Update guess for the next integration step
            rhoT_guess_metastable = [state_meta.rho, state_meta.T]
            dhdp = np.atleast_1d(efficiency / state_meta["rho"])
            return dhdp, state_meta

        elif calculation_type == "blending":
            # Compute equilibrium state using CoolProp solver
            state_eq = fluid.get_state(
                input_type=props.HmassP_INPUTS,
                prop_1=h,
                prop_2=p,
                generalize_quality=True,
                supersaturation=True,
            )

            # Compute metastable state using custom solver
            state_meta = fluid.get_state_metastable(
                prop_1="h",
                prop_1_value=h,
                prop_2="p",
                prop_2_value=p,
                rhoT_guess=rhoT_guess_metastable,
                generalize_quality=False,
                supersaturation=True,
                solver_algorithm=HEOS_solver,
                solver_tolerance=HEOS_tolerance,
                solver_max_iterations=HEOS_max_iter,
                print_convergence=HEOS_print_convergence,
            )

            # Update guess for the next integration step
            rhoT_guess_metastable = [state_meta.rho, state_meta.T]

            # Compute blended state
            state_blend = props.blend_properties(
                phase_change=phase_change_type,
                props_equilibrium=state_eq,
                props_metastable=state_meta,
                blending_variable="Q",
                blending_onset=blending_onset,
                blending_width=blending_width,
            )

            # Compute right-hand-side of the ODE
            dhdp = np.atleast_1d(efficiency / state_blend["rho"])
            return dhdp, props.FluidState(state_blend, fluid.name)

        else:
            raise ValueError(
                f"Invalid calculation_type='{calculation_type}'. "
                f"Valid options are: {', '.join(PROPERTY_CALCULATION_TYPES)}."
            )

    # Check if the provided ODE_solver is valid
    valid_solvers_heos = psv.nonlinear_system.SOLVER_OPTIONS
    if HEOS_solver not in valid_solvers_heos:
        error_message = (
            f"Invalid HEOS solver '{HEOS_solver}' provided. "
            f"Valid solvers are: {', '.join(valid_solvers_heos)}. "
        )
        raise ValueError(error_message)

    # Check if the provided ODE_solver is valid
    valid_solvers_ode = list(ODE_METHODS.keys())
    if ODE_solver not in valid_solvers_ode:
        error_message = (
            f"Invalid ODE solver '{ODE_solver}' provided. "
            f"Valid solver are: {', '.join(valid_solvers_ode)}. "
            "Recommended solvers: 'BDF', 'LSODA' or 'Radau' stiff problems involving equilibrium property calculations or blending calculations with a narrow blending width. 'RK45' can be used for non-stiff problems with a wide blending width."
        )
        raise ValueError(error_message)

    # Solve polytropic expansion differential equation
    ode_sol = solve_ivp(
        lambda p, h: odefun(p, h)[0],  # Get only first output
        [p_in, p_out],
        [state_in.h],
        method=ODE_solver,
        rtol=ODE_tolerance,
        atol=ODE_tolerance,
    )
    if not ode_sol.success:
        raise Exception(ode_sol.message)

    # Postprocess solution
    states = utils.postprocess_ode(ode_sol.t, ode_sol.y, odefun)

    return states, ode_sol



def _ensure_data_available(method):
    """
    Decorator to ensure that poly_handles and poly_breakpoints are available before executing a method.
    """

    @wraps(method)
    def wrapper(self, *args, **kwargs):
        if not self.poly_handles or not self.poly_breakpoints:
            raise ValueError(
                "Polynomials or breakpoints are not initialized. Please fit the model first."
            )
        return method(self, *args, **kwargs)

    return wrapper


class PolynomialFitter:
    """
    Fits polynomials to the thermodynamic properties of a fluid across various states.

    polynomial_degree : int
        Degree of the polynomials to fit. Default is 8.

        .. note::

            When `calculation_type` is ``blending`` the degree of the polynomial in the blending region is set to 4 to achieve sufficient accuracy while preventing numerical round-off errors associated with single-precision arithmetic in CFD solvers.

    polynomial_format : str, optional
        Type of polynomial representation (``horner`` or ``standard``). Default is 'horner'.
    polynomial_variables : list of str
        A list of variable names to fit polynomials to, such as 'density', 'viscosity',
        'speed_sound', 'void_fraction', 'vapor_quality'.
    output_dir : str
        The directory where output will be saved.

    """

    def __init__(
        self,
        states,
        variables,
        degree,
        calculation_type,
        output_dir="barotropic_model",
    ):

        # Rename arguments
        self.states = states
        self.variables = variables
        self.degree = degree
        self.calculation_type = calculation_type
        self.output_dir_default = output_dir

        # Initialize variables
        self.poly_handles = {}
        self.poly_breakpoints = None
        self.p_in = self.states["p"][0]
        self.p_out = self.states["p"][-1]
        self.p_in_scaled = self.p_in / self.p_in
        self.p_out_scaled = self.p_out / self.p_in
        self.plot_labels = {
            "density": "Density (kg/m$^3$)",
            "viscosity": "Viscosity (Pa·s)",
            "speed_sound": "Speed of sound (m/s)",
            "void_fraction": "Void fraction",
            "vapor_quality": "Vapor quality",
        }

    def fit_polynomials(self):
        """
        Fits polynomials to the data obtained from the ODE solution based on the specified calculation type.
        """
        if self.calculation_type == "equilibrium":
            self._fit_equilibrium()
        elif self.calculation_type == "blending":
            self._fit_blending()
        else:
            raise ValueError("Unsupported calculation type")

    def _fit_equilibrium(self):
        """
        Fits polynomials for calculation_type="equilibrium".

        This method scales pressure, identifies the phase transition, and fits polynomials
        for the specified variables based on single-phase and two-phase regions.
        """
        # Scale pressure by inlet pressure to improve polynomial conditioning
        p_scaled = self.states["p"] / self.p_in

        # Determine points within the two-phase region
        eps = 1e-9
        mask_1phase = np.abs(self.states["supersaturation_degree"]) > eps
        mask_2phase = np.abs(self.states["supersaturation_degree"]) <= eps

        # Determine the phase change pressure if it exists and the breakpoints for the different polynomial segments
        if mask_2phase.any():
            p_transition = p_scaled[np.where(mask_2phase)[0][0]]
            self.poly_breakpoints = [self.p_in_scaled, p_transition, self.p_out_scaled]
        else:
            self.poly_breakpoints = [self.p_in_scaled, self.p_out_scaled]

        # Fit polynomials to data
        for var in self.variables:
            self.poly_handles[var] = []

            if mask_1phase.any():  # Single-phase data
                y_1p = np.array(self.states[var])[mask_1phase]
                p_1p = p_scaled[mask_1phase]
                poly_1p = Polynomial.fit(p_1p, y_1p, deg=self.degree).convert()
                self.poly_handles[var].append(poly_1p)

            if mask_2phase.any():  # Two-phase data
                y_2p = np.array(self.states[var])[mask_2phase]
                p_2p = p_scaled[mask_2phase]
                poly_2p = Polynomial.fit(p_2p, y_2p, deg=self.degree).convert()

                # # Ensure continuity by adjusting the first coefficient
                # y_1p_breakpoint = poly_1p(p_transition)
                # y_2p_breakpoint = poly_2p(p_transition)
                # coeffs_2p = poly_2p.coef
                # coeffs_2p[0] -= y_2p_breakpoint - y_1p_breakpoint
                # poly_2p = Polynomial(coeffs_2p)
                self.poly_handles[var].append(poly_2p)

    def _fit_blending(self):
        """
        Fits polynomials for calculation_type="blending".

        This method identifies different regions based on the blending
        parameter "x" and fits polynomials to ensure continuity across regions.
        """

        # Get the scaled pressure values
        p = self.states["p"] / self.p_in

        # Define masks for different regions based on states["x"]
        mask_region_1 = self.states["x"] > 1
        mask_region_2 = (self.states["x"] >= 0) & (self.states["x"] <= 1)
        mask_region_3 = self.states["x"] < 0

        # Determine breakpoints in terms of scaled pressure
        p_region_2_start = p[mask_region_2][0]
        p_region_2_end = p[mask_region_2][-1]
        self.poly_breakpoints = [
            self.p_in_scaled,
            p_region_2_start,
            p_region_2_end,
            self.p_out_scaled,
        ]

        # Fit polynomials to data
        for var in self.variables:
            self.poly_handles[var] = []

            # Fit polynomial for Region 1 (x > 1)
            y_1 = np.array(self.states[var])[mask_region_1]
            p_1 = p[mask_region_1]
            poly_1 = Polynomial.fit(p_1, y_1, deg=self.degree).convert()
            self.poly_handles[var].append(poly_1)

            # Fit polynomial for Region 2 (0 <= x <= 1)
            # TODO Hardcoded degree 4 to prevent numerical noise
            y_2 = np.array(self.states[var])[mask_region_2]
            p_2 = p[mask_region_2]
            poly_2 = Polynomial.fit(p_2, y_2, deg=4).convert()  

            # Ensure continuity by adjusting the first coefficient
            y_1_end = poly_1(p_region_2_start)
            y_2_start = poly_2(p_region_2_start)
            coeffs_2 = poly_2.coef
            coeffs_2[0] -= y_2_start - y_1_end
            poly_2 = Polynomial(coeffs_2)
            self.poly_handles[var].append(poly_2)

            # Fit polynomial for Region 3 (x < 0)
            y_3 = np.array(self.states[var])[mask_region_3]
            p_3 = p[mask_region_3]
            poly_3 = Polynomial.fit(p_3, y_3, deg=self.degree).convert()

            # Ensure continuity by adjusting the first coefficient
            y_2_end = poly_2(p_region_2_end)
            y_3_start = poly_3(p_region_2_end)
            coeffs_3 = poly_3.coef
            coeffs_3[0] -= y_3_start - y_2_end
            poly_3 = Polynomial(coeffs_3)
            self.poly_handles[var].append(poly_3)

    @_ensure_data_available
    def evaluate_polynomial(self, p, variable):
        r"""
        Evaluates the polynomials for a given variable at a specified pressure values.

        The function evaluates the polynomial values at a given physical pressure `p`
        for a variable (e.g., density). It automatically determines the correct branch
        of the piecewise polynomial based on the breakpoints, and applies safeguards for 
        pressures outside the limits to ensure sensible values and continuity.

        The evaluation follows a piecewise definition depending on the range of :math:`\hat{p}=p/p_{\text{in}}`:

        .. math::

            \phi(\hat{p}) = 
            \begin{cases}
            a_1 \, e^\left(\frac{\hat{p} - \hat{p}_{\text{out}}}{a_2}\right) & \text{ for } & \hat{p} < \hat{p}_{\text{out}} \\
            \sum_{i=0}^{d} b_{i,1} \, \hat{p}^i & \text{ for }  & \hat{p}_{\text{ out}} \leq \hat{p} \leq \hat{p}_{1} \\
            & \;\; \vdots  & \\
            \sum_{i=0}^{d} b_{i,n} \, \hat{p}^i & \text{ for }  & \hat{p}_{n} \leq \hat{p} \leq \hat{p}_{\text{in}} \\
            c_1 + c_2 \left(\hat{p} - \hat{p}_{\text{in}}\right) & \text{ for }  & \hat{p} > \hat{p}_{\text{in}}
            \end{cases}

        where:

        - :math:`\phi(p)` is the variable being fitted (e.g., density, viscosity).
        - :math:`a_1` and :math:`a_2` are constants of the exponential function used to extrapolate the properties below the outlet pressure. The numerical values are determined to match the polynomial value and its first derivative at the endpoint :math:`\hat{p}=\hat{p}_{\text{out}}`.
        - :math:`b_{i,j}` is the :math:`i`-th polynomial coefficient of the :math:`j`-th polynomial segment, where :math:`d` is the degree of the polynomials, :math:`n` is the number of breakpoints, and :math:`n+1` is the number of polynomial segments. Additionally, :math:`[\hat{p}_{\text{out}},\, \hat{p}_{1},\,\ldots,\, \hat{p}_{n},\,\hat{p}_{\text{in}}]` are normalized pressures at the breakpoints.
        - :math:`c_1` and :math:`c_2` are constants of the linear function used to extrapolate the properties above the inlet pressure. The numerical values are determined to match the polynomial value and its first derivative at the endpoint :math:`\hat{p}=\hat{p}_{\text{in}}`.

        .. note::

            The safeguards for pressures outside the specified limits help to avoid numerical issues by applying 
            exponential decay for pressures below :math:`p_{\text{out}}` and linear extrapolation for pressures above 
            :math:`p_{\text{in}}`. These measures prevent properties from becoming negative or excessively large, which 
            can occur with very high or low (even negative) pressures during internal iterations of CFD solvers.


        Parameters
        ----------
        p : array-like
            The physical pressure values at which to evaluate the polynomials. 
            Given in Pascals (Pa).
        variable : str
            The variable for which to evaluate the polynomial, such as 'density', 'viscosity', 
            'speed_of_sound', or 'void_fraction'.

        Returns
        -------
        var_values : ndarray
            The evaluated values of the variable at the given physical pressures.

        """

        # Note:
        # The argument to the polynomials in self.poly_handles() should be normalized pressures (not physical pressures)
        # The values in self.poly_breakpoints are normalized pressures (not physical pressures)

        # Convert input to array if not already
        p = np.atleast_1d(p)
        var_values = np.zeros_like(p, dtype=float)

        # Define scaled pressures for internal calculations
        p_scaled = p / self.p_in
        bps = np.array(self.poly_breakpoints)
        if variable in self.poly_handles.keys():
            for i, poly in enumerate(self.poly_handles.get(variable, [])):
                mask = (p_scaled <= bps[i]) & (p_scaled > bps[i + 1])
                if np.any(mask):
                    var_values[mask] = poly(p_scaled[mask])
        else:
            raise ValueError(f"No polynomials found for variable '{variable}'")

        # Safeguard for pressures lower than the lowest breakpoint (exponential decay)
        mask_lower = p_scaled <= self.p_out_scaled
        if np.any(mask_lower):
            poly_last = self.poly_handles[variable][-1]
            a1 = poly_last(self.p_out_scaled)
            a2 = a1 / poly_last.deriv(m=1)(self.p_out_scaled)
            exp_values = a1 * np.exp((p_scaled[mask_lower] - self.p_out_scaled) / a2)
            var_values[mask_lower] = exp_values

        # Safeguard for pressures higher than the highest breakpoint (linear extrapolation)
        mask_upper = p_scaled > self.p_in_scaled
        if np.any(mask_upper):
            poly_first = self.poly_handles[variable][0]
            b1 = poly_first(self.p_in_scaled)
            b2 = poly_first.deriv(m=1)(self.p_in_scaled)
            extrap = b1 + b2 * (p_scaled[mask_upper] - self.p_in_scaled)
            var_values[mask_upper] = extrap

        return var_values

    @_ensure_data_available
    def plot_polynomial(self, var, p_eval=None, showfig=True, savefig=False, output_dir=None):
        """
        Plots the polynomials for the specified variable. If the pressure values for plotting are not provided, a suitable range is generated internally.

        Parameters
        ----------
        var : str
            The variable for which to plot the polynomial.
        p_eval : array-like, optional
            The pressure values at which to plot the polynomial. If not provided, automatic values are generated.
        showfig : bool, optional
            If True, displays the plot. Default is True.
        savefig : bool, optional
            If True, saves the plot to a file. Default is False.
        output_dir : str, optional
            The directory where the plot will be saved if `savefig` is True.
            If not specified, uses the default output directory of the class.

        Returns
        -------
        fig : matplotlib.figure.Figure
            The figure object containing the plotted data.
        """

        # If output_dir is not given, set it to the default directory
        output_dir = output_dir or self.output_dir_default
        if not os.path.isdir(output_dir):
            os.makedirs(output_dir)

        # Create a figure and axis
        fig, ax = plt.subplots(figsize=(6, 5))

        # Set axis labels
        ax.set_xlabel("Pressure (Pa)")
        ax.set_ylabel(self.plot_labels.get(var, var))

        # Plot specified values or segments between breakpoints
        if p_eval is not None:

            # Plot polynomial at specified pressures
            prop_eval = self.evaluate_polynomial(p_eval, var)
            ax.plot(p_eval, prop_eval, color=COLORS_MATLAB[0])

        else:

            # Iterate over the breakpoints and plot the segments
            extrap_lower = [0.8 * self.poly_breakpoints[-1]]
            extrap_upper = [1.2 * self.poly_breakpoints[0]]
            breakpoints = extrap_upper + self.poly_breakpoints + extrap_lower
            for i in range(len(breakpoints) - 1):

                # Generate segments between breakpoints
                bp1 = breakpoints[i]
                bp2 = breakpoints[i + 1]
                p_segment = np.linspace(bp1, bp2, 100) * self.p_in

                # Evaluate the polynomial over the segment
                prop_eval = self.evaluate_polynomial(p_segment, var)
                ax.plot(p_segment, prop_eval, color=COLORS_MATLAB[i])

                # Plot points just before breakpoints to check continuity
                eps = 1e-12
                x_minus = (bp2 - eps) * self.p_in
                y_minus = self.evaluate_polynomial(x_minus, var)
                ax.plot(x_minus, y_minus, marker="o", color=COLORS_MATLAB[i])

                # Plot points just after breakpoints to check continuity
                x_plus = (bp2 + eps) * self.p_in
                y_plus = self.evaluate_polynomial(x_plus, var)
                ax.plot(x_plus, y_plus, marker="o", color=COLORS_MATLAB[i])

        plt.tight_layout(pad=1)
        
        if savefig:
            file_path = os.path.join(output_dir, f"barotropic_model_{var}")
            graphics.savefig_in_formats(fig, file_path)

        if not showfig:
            plt.close(fig)

        return fig

    @_ensure_data_available
    def plot_polynomial_and_error(self, var, showfig=True, savefig=False, output_dir=None):
        r"""
        Plots the barotropic polynomials and original data points from the ODE solution to illustrate
        the quality of the fit. Additionally, plots the relative error between the polynomial and the data.

        Parameters
        ----------
        var : str
            The variable to plot (e.g., 'density', 'viscosity', etc.).
        showfig : bool, optional
            If True, displays the plot. Default is True.
        savefig : bool, optional
            If True, saves the plot to a file. Default is False.
        output_dir : str, optional
            The directory where the plot will be saved if `savefig` is True.
            If not specified, uses the default output directory of the class.

        Returns
        -------
        fig : matplotlib.figure.Figure
            The figure object containing the plotted data and error.
        """
        # If output_dir is not given, set it to the default directory
        output_dir = output_dir or self.output_dir_default
        if not os.path.isdir(output_dir):
            os.makedirs(output_dir)

        # Create a figure with two vertically stacked subplots
        fig, (ax1, ax2) = plt.subplots(2, 1, sharex=True, figsize=(6, 5))

        # Mapping variable names to axis labels (variable name if not found)
        x_label = "Pressure (Pa)"
        y_label = self.plot_labels.get(var, var)

        # Generate pressure values and evaluate the variable
        pressures = self.states["p"]
        prop_poly = self.evaluate_polynomial(pressures, var)
        prop_states = self.states[var]

        # Compute the relative error
        relative_error = 100 * np.abs(prop_poly - prop_states) / np.abs(prop_states)

        # Plot the variable from polynomials and states in the upper subplot
        ax1.plot(
            pressures,
            prop_poly,
            label=f"Polynomial fit",
            linestyle="-",
            color=COLORS_MATLAB[1],
        )
        ax1.plot(
            pressures,
            prop_states,
            label=f"Original values",
            marker="o",
            markersize=2.5,
            linestyle="none",
            color=COLORS_MATLAB[0],
        )
        ax1.set_ylabel(y_label)
        ax1.legend(loc="lower right")

        # Plot the relative error in the lower subplot
        ax2.plot(
            pressures,
            relative_error,
            label="Relative Error",
            linestyle="-",
            markersize=2.5,
            color=COLORS_MATLAB[1],
        )
        ax2.set_xlabel(x_label)
        ax2.set_ylabel("Relative error (%)")
        plt.tight_layout(pad=1)

        if savefig:
            file_path = os.path.join(output_dir, f"barotropic_model_error_{var}")
            graphics.savefig_in_formats(fig, file_path)

        if not showfig:
            plt.close(fig)
            
        return fig


class ExpressionExporter:
    """
    Exports expressions describing the barotropic model for use within CFD solvers.

    Parameters
    ----------
    poly_fitter : PolynomialFitter
        An instance of PolynomialFitter containing the polynomial coefficients and breakpoints.
    poly_format : str, optional
        Type of polynomial representation ('horner' or 'standard'). Default is 'horner'.
    """

    def __init__(self, poly_fitter, poly_format="horner"):
        # Rename arguments
        self.poly_fitter = poly_fitter
        self.poly_format = poly_format

        # Define the units of the fluid properties
        self.units = {
            "density": "kg/m^3",
            "viscosity": "Pa*s",
            "speed_of_sound": "m/s",
            "void_fraction": "none",
            "vapor_quality": "none",
        }

        # Define syntax specifics for different solvers
        self.solver_syntax = {
            "fluent": {"if": "IF", "pressure": "AbsolutePressure"},
            "cfx": {"if": "if", "pressure": "Absolute Pressure"},
        }

        # Format of the numbers in the exported expressions
        self.num_format = "0.16e"

        # Define default output directory
        self.output_dir_default = poly_fitter.output_dir_default


    def export_expressions_fluent(self, output_dir=None):
            """
            Exports the barotropic model polynomials as expressions for use in Fluent CFD software.

            Parameters
            ----------
            output_dir : str, optional
                The directory where the expressions will be saved. It uses a default directory if not provided.
            """
            self._export_expressions(solver="fluent", output_dir=output_dir)

    def export_expressions_cfx(self, output_dir=None):
        """
        Exports the barotropic model polynomials as expressions for use in CFX CFD software.

        Parameters
        ----------
        output_dir : str, optional
            The directory where the expressions will be saved. It uses a default directory if not provided.
        """
        self._export_expressions(solver="cfx", output_dir=output_dir)


    def _export_expressions(self, solver, output_dir=None):
        """
        A generic method to generate expressions for a CFD solver, given specific keywords.

        Parameters
        ----------
        solver : str
            The solver for which to generate expressions ('fluent' or 'cfx').
        output_dir : str, optional
            The directory where the expressions will be saved. It uses a default directory if not provided.
        """
        # If output_dir is not given, set it to the default directory
        output_dir = output_dir or self.output_dir_default
        if not os.path.isdir(output_dir):
            os.makedirs(output_dir)

        # Define syntax based on solver
        self.syntax = self.solver_syntax[solver]

        # Create expressions
        file_path = os.path.join(output_dir, f"{solver}_expressions.txt")
        with open(file_path, "wt") as file:
            # File header
            file.write(f"{solver.upper()} expressions for barotropic properties\n")
            file.write(f"Creation datetime: {datetime.datetime.now()}\n\n")

            # Generate expressions for each property
            for var in self.poly_fitter.variables:
                if var in self.units:
                    unit = self.units[var]
                    expressions = self.generate_expressions(var, unit)
                    file.write(f"barotropic_model_{var}\n")
                    file.write(f"{expressions}\n\n")


    def generate_expressions(self, var, unit):
        """
        Generates a piecewise polynomial expression with extrapolation safeguards for a given variable.

        Parameters
        ----------
        var : str
            The variable name for which to generate the expression.
        unit : str
            The unit of the variable.

        Returns
        -------
        str
            The piecewise polynomial expression with extrapolation safeguards.
        """

        # Retrieve polynomial coefficients
        p_in = self.poly_fitter.p_in
        p_in_scaled = self.poly_fitter.p_in_scaled
        p_out_scaled = self.poly_fitter.p_out_scaled
        polynomials = self.poly_fitter.poly_handles[var]
        breakpoints = [p_in * bp for bp in self.poly_fitter.poly_breakpoints]

        # Initialize expressions
        expression = self.polynomial_expression(polynomials[0].coef)
        for i in range(1, len(polynomials)):
            new_expression = self.polynomial_expression(polynomials[i].coef)
            expression = self.if_expression(expression, new_expression, breakpoints[i])

        # Safeguard for pressures lower than p_out (Exponential decay)
        poly_last = polynomials[-1]
        a1 = poly_last(p_out_scaled)
        a2 = a1 / poly_last.deriv(m=1)(p_out_scaled)
        exp_safeguard = f"{a1:{self.num_format}} * exp(({self.syntax['pressure']} / {p_in:{self.num_format}} [Pa] - {p_out_scaled}) / {a2:{self.num_format}})"
        expression = self.if_expression(expression, exp_safeguard, breakpoints[-1])

        # Safeguard for pressures higher than p_in (Linear extrapolation)
        poly_first = polynomials[0]
        c1 = poly_first(p_in_scaled)
        c2 = poly_first.deriv(m=1)(p_in_scaled)
        lin_extrap = f"{c1:{self.num_format}} + {c2:{self.num_format}} * ({self.syntax['pressure']} / {p_in:{self.num_format}} [Pa] - {p_in_scaled:{self.num_format}})"
        expression = self.if_expression(lin_extrap, expression, breakpoints[0])

        # Add fluid property unit
        if unit != "none":
            expression = f"{expression} * 1 [{unit}]"
        return expression

    def polynomial_expression(self, coefficients):
            """
            Generates a polynomial expression string.

            Parameters
            ----------
            coefficients : array-like
                Coefficients of the polynomial.

            Returns
            -------
            str
                The polynomial expression string.
            """
            p_scale = self.poly_fitter.p_in
            if self.poly_format == "horner":
                coefficients = list(reversed(coefficients))
                polynomial_string = f"{coefficients[0]:{self.num_format}}"
                for i in range(1, len(coefficients)):
                    polynomial_string = f"{coefficients[i]:{self.num_format}} + ({self.syntax['pressure']} / {p_scale:{self.num_format}} [Pa]) *\n({polynomial_string})"
            elif self.poly_format == "standard":
                terms = []
                for i in range(len(coefficients)):
                    summand = f"{coefficients[i]:{self.num_format}} * ({self.syntax['pressure']} / {p_scale:{self.num_format}} [Pa])^{i}"
                    if i < len(coefficients) - 1:
                        summand += " +"  
                    terms.append(summand)
                polynomial_string = " \n".join(terms)
            else:
                raise ValueError("The polynomial format must be 'horner' or 'standard'")
            return polynomial_string

    def if_expression(self, expression_1, expression_2, transition_pressure):
            """
            Generates an IF expression string for piecewise functions.

            Parameters
            ----------
            expression_1 : str
                The expression for the condition being true.
            expression_2 : str
                The expression for the condition being false.
            transition_pressure : float
                The transition pressure where the expressions change.

            Returns
            -------
            str
                The IF expression string.
            """
            return f"{self.syntax['if']}({self.syntax['pressure']} >= {transition_pressure:{self.num_format}} [Pa], \n{expression_1}, \n{expression_2})"






import numpy as np
from scipy.optimize._numdiff import approx_derivative

DERIVATIVE_METHODS = [
    "2-point",
    "3-point",
    "Complex step method",
    "forward_finite_differences",
    "central_finite_differences",
    "complex_step",
]


def approx_gradient(
    function_handle, x, f0=None, abs_step=None, method="central_finite_differences"
):
    """
    Approximate the derivatives of a given function at a point using various differentiation methods.

    Parameters
    ----------
    function_handle : callable
        The function for which derivatives are to be approximated.
    x : array_like
        The point at which derivatives are to be approximated.
    abs_step : float or np.ndarray with the same size as x, optional
        Step size for finite difference methods. By default uses a suitable value for each method.
    method : str, optional
        The method to use for differentiation. Available options are:
        - 'forward_finite_differences' or '2-point': Forward finite differences.
        - 'central_finite_differences' or '3-point': Central finite differences.
        - 'complex_step' or 'cs': Complex step method.

        Defaults to 'central_finite_differences'.

    Returns
    -------
    numpy.ndarray
        The approximated derivatives of the function at the given point `x`.

    Raises
    ------
    ValueError
        If an invalid method is provided.
    """
    if method == "forward_finite_differences" or method == "2-point":
        return forward_finite_differences(function_handle, x, abs_step, f0)
    elif method == "central_finite_differences" or method == "3-point":
        return central_finite_differences(function_handle, x, abs_step)
    elif method == "complex_step" or method == "cs":
        return complex_step_derivative(function_handle, x, abs_step)
    else:
        raise ValueError(
            f"Invalid method '{method}' provided. Available methods: 'forward_finite_differences', "
            "'central_finite_differences', 'complex_step'."
        )


def forward_finite_differences(function_handle, x, abs_step=None, f0=None):
    """
    Gradient approximation by forward finite differences.

    Parameters
    ----------
    function_handle : callable
        The function for which the derivative is calculated.
    x : np.ndarray
        The point at which the derivative is calculated.
    abs_step : float or np.ndarray with the same size as x, optional
        The step size for finite differences. Default is square root of machine epsilon.

    Returns
    -------
    np.ndarray
        The gradient of the function at point x.
    """

    # Default step size
    if abs_step is None:
        # abs_step = np.finfo(float).eps ** (1 / 2)
        abs_step = np.finfo(float).eps ** (1 / 2)

    # Ensure abs_step is an array of the same shape as x
    abs_step = np.broadcast_to(abs_step, x.shape)

    # Avoid one function evaluation if f0 is provided
    if f0 is None:
        f0 = np.atleast_1d(function_handle(x))
    else:
        f0 = np.atleast_1d(f0)

    # Compute gradient
    m = np.size(f0)
    n = np.size(x)
    df = np.zeros((m, n))
    for i in range(n):
        x_step = np.copy(x)
        x_step[i] += abs_step[i]
        df[:, i] = (np.atleast_1d(function_handle(x_step)) - f0) / abs_step[i]

    return df.squeeze()


def central_finite_differences(function_handle, x, abs_step=None):
    """
    Gradient approximation by central finite differences.

    Parameters
    ----------
    function_handle : callable
        The function for which the derivative is calculated.
    x : np.ndarray
        The point at which the derivative is calculated.
    abs_step : float or np.ndarray with the same size as x, optional
        The step size for finite differences. Default is cobic root of machine epsilon.

    Returns
    -------
    np.ndarray
        The gradient of the function at point x.
    """

    # Default step size
    if abs_step is None:
        abs_step = np.finfo(float).eps ** (1 / 3)

    # Ensure abs_step is an array of the same shape as x
    abs_step = np.broadcast_to(abs_step, x.shape)

    # Compute gradient
    f0 = np.atleast_1d(function_handle(x))
    m = np.size(f0)
    n = np.size(x)
    df = np.zeros((m, n))
    for i in range(n):
        x_step_forward = np.copy(x)
        x_step_backward = np.copy(x)
        x_step_forward[i] += abs_step[i]
        x_step_backward[i] -= abs_step[i]
        df[:, i] = (
            np.atleast_1d(function_handle(x_step_forward))
            - np.atleast_1d(function_handle(x_step_backward))
        ) / (2 * abs_step[i])

    return df.squeeze()


def complex_step_derivative(function_handle, x, abs_step=None):
    """
    Gradient approximation using the complex step method.

    Parameters
    ----------
    function_handle : callable
        The function for which the derivative is calculated.
    x : np.ndarray
        The point at which the derivative is calculated.
    abs_step : float or np.ndarray with the same size as x, optional
        The step size for the complex step. Default is machine epsilon.

    Returns
    -------
    np.ndarray
        The gradient of the function at point x.
    """

    # Default step size
    if abs_step is None:
        abs_step = np.finfo(float).eps

    # Ensure abs_step is an array of the same shape as x
    abs_step = np.broadcast_to(abs_step, x.shape)

    # Compute gradient
    f0 = np.atleast_1d(function_handle(x))
    m = np.size(f0)
    n = np.size(x)
    df = np.zeros((m, n), dtype=np.complex128)
    for i in range(n):
        x_step = x.astype(np.complex128)
        x_step[i] += abs_step[i] * 1j
        df[:, i] = np.imag(np.atleast_1d(function_handle(x_step))) / abs_step[i]

    return df.squeeze().real


def approx_jacobian_hessians(f, x, abs_step=1e-5, lower_triangular=True):
    """
    Calculate the Hessian matrices for each component of a vector-valued function using finite differences.

    Parameters
    ----------
    f : callable
        The function for which to find the Hessian matrices. It must take a
        single argument which is a numpy array and can return a scalar or a numpy array.
    x : numpy array
        The point at which the Hessian matrices are calculated.
    abs_step : float, optional
        The step size for the finite differences, default is 1e-5.
    lower_triangular : bool, optional
        If True, the Hessians are returned in a lower triangular form suitable for Pygmo, default is True.

    Returns
    -------
    Hessians : numpy array
        A tensor where each slice along the first dimension corresponds to the Hessian matrix
        of each component of the function f at x. If f returns a scalar, the first dimension size is 1.
        If lower_triangular is True, the Hessians are returned in a lower triangular form.
    """
    x = np.asarray(x)
    f0 = np.atleast_1d(f(x))
    m = len(f0)
    n = len(x)
    Hessians = np.zeros((m, n, n))

    hh = np.eye(n) * abs_step

    for i in range(n):
        for j in range(i, n):
            f_ij = np.atleast_1d(f(x + hh[i] + hh[j]))
            f_i = np.atleast_1d(f(x + hh[i]))
            f_j = np.atleast_1d(f(x + hh[j]))

            Hessians[:, i, j] = (f_ij - f_i - f_j + f0) / abs_step**2
            if i != j:  # Symmetry
                Hessians[:, j, i] = Hessians[:, i, j]

    # Reformat to lower-triangular form suitable for Pygmo
    if lower_triangular:
        k = (n * (n + 1)) // 2
        Hessians_L = np.zeros((m, k))
        for i in range(m):
            Hessians_L[i] = Hessians[i][np.tril_indices(n)]
        Hessians = Hessians_L

    return Hessians



import os
import time
import copy
import logging
import warnings
import numpy as np
import matplotlib.pyplot as plt

from datetime import datetime
from abc import ABC, abstractmethod
from matplotlib.ticker import MultipleLocator
from scipy.optimize import root

from . import numerical_differentiation
from .pysolver_utilities import savefig_in_formats

SOLVER_OPTIONS = ["hybr", "lm"]


class NonlinearSystemSolver:
    r"""
    Solver class for nonlinear systems of equations.

    The solver is designed to handle system of nonlinear equations of the form:

    .. math::

        F(x) = 0

    where :math:`F: \mathbb{R}^n \rightarrow \mathbb{R}^n` is a vector-valued
    function of the vector :math:`x`.

    The class interfaces with the `root` method from `scipy.optimize` to solve the equations and provides a structured framework for initialization, solution monitoring, and post-processing.

    Parameters
    ----------
    problem : NonlinearSystemProblem
        An instance of a problem defining the system of equations to be solved.
    method : str, optional
        Method to be used by scipy's root for solving the nonlinear system. Available solvers are:

        - :code:`hybr`:  Uses MINPACK's 'hybrd' method, which is is a modification of Powell's hybrid method.
        - :code:`lm`: The Levenberg-Marquardt method, which blends the steepest descent and the Gauss-Newton methods.

        The choice between :code:`hybr` and :code:`lm` largely depends on the specifics of the problem at hand.
        The :code:`hybr` usually requires less gradient evaluations and it is often faster when analytic gradients are not available.
        It is advisable to experiment with both methods to determine the most appropriate choice for a given problem.

        Defaults to 'hybr'.
    tol : float, optional
        Tolerance for the solver termination. Defaults to 1e-9.
    max_iter : integer, optional
        Maximum number of function evaluations for the solver termination. Defaults to 100.
    options : dict, optional
        Additional options to be passed to scipy's root.
    derivative_method : str, optional
        Finite difference method to be used when the problem Jacobian is not provided. Defaults to '2-point'
    derivative_abs_step : float, optional
        Finite difference absolute step size to be used when the problem Jacobian is not provided. Defaults to 1e-6
    print_convergence : bool, optional
        If True, displays the convergence progress. Defaults to True.
    plot_convergence : bool, optional
        If True, plots the convergence progress. Defaults to False.
    logger : logging.Logger, optional
        Logger object to which logging messages will be directed. Defaults to None.
    update_on : str, optional
        Specifies if the convergence report should be updated on a new function evaluations ("function") or on gradient evaluations ("gradient"). Defaults to "function".


    Methods
    -------
    solve(x0)
        Solve the system of nonlinear equations using the specified initial guess `x0`.
    residual(x)
        Evaluate the vector of residuals of the at a given point `x`.
    gradient(x)
        Evaluate the Jacobian of the system at a given point `x`.
    print_convergence_history()
        Print the convergence history of the problem.
    plot_convergence_history()
        Plot the convergence history.

    """

    def __init__(
        self,
        problem,
        method="hybr",
        tolerance=1e-6,
        max_iterations=100,
        options={},
        derivative_method="2-point",
        derivative_abs_step=1e-6,
        print_convergence=True,
        plot_convergence=False,
        plot_scale="log",
        logger=None,
        update_on="function",
        callback_func=None,
    ):
        # Initialize class variables
        self.problem = problem
        self.display = print_convergence
        self.plot = plot_convergence
        self.plot_scale = plot_scale
        self.logger = logger
        self.method = method
        self.tol = tolerance
        self.maxiter = max_iterations
        self.options = copy.deepcopy(options) if options else {}
        self.derivative_method = derivative_method
        self.derivative_abs_step = derivative_abs_step
        self.callback_func = callback_func
        self.callback_func_call_count = 0

        # Define the maximun number of iterations
        if method == "hybr":
            self.options["maxfev"] = self.maxiter
        elif method == "lm":
            self.options["maxiter"] = self.maxiter
        else:
            raise ValueError(
                f"Invalid solver. Available options: {', '.join(SOLVER_OPTIONS)}"
            )

        # Check for logger validity
        if self.logger is not None:
            if not isinstance(self.logger, logging.Logger):
                raise ValueError(
                    "The provided logger is not a valid logging.Logger instance."
                )

        # Check for valid display_on value
        self.update_on = update_on
        if update_on not in ["function", "gradient"]:
            raise ValueError(
                "Invalid value for 'update_on'. It should be either 'function' or 'gradient'."
            )

        # Initialize variables for convergence report
        self.f_final = None
        self.x_final = None
        self.x_last = None
        self.residuals_last = None
        self.grad_count = 0
        self.func_count = 0
        self.func_count_tot = 0
        self.success = None
        self.message = None
        self.solution_report = []
        self.elapsed_time = None
        self.include_solution_in_footer = False
        self.convergence_history = {
            "grad_count": [],
            "func_count": [],
            "func_count_total": [],
            "norm_residual": [],
            "norm_step": [],
        }

        # Initialize convergence plot
        # TODO ??? (convergence_history must be initialized before)
        if self.plot:
            self._plot_callback(initialize=True)

    def solve(self, x0):
        """
        Solve the nonlinear system using the specified solver.

        Parameters
        ----------
        x0 : array-like
            Initial guess for the solution of the nonlinear system.

        Returns
        -------
        RootResults
            A result object from scipy's root detailing the outcome of the solution process.

        """
        # Get start datetime
        self.start_datetime = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")

        # Start timing with high-resolution timer
        start_time = time.perf_counter()

        # Print report header
        self._write_header()

        # Solve the root finding problem
        solution = root(
            self.residual,
            x0,
            jac=self.gradient,
            method=self.method,
            tol=self.tol,
            options=self.options,
        )

        # Save solution
        self.x_final = copy.deepcopy(self.x_last)
        self.f_final = self.residual(self.x_final)
        self.success = solution.success
        self.message = solution.message

        # Check if solver actually converged
        norm_residual = np.linalg.norm(self.f_final)
        if self.success and norm_residual > 10*self.tol:
            self.success = False
            self.message = f"The equation solver returned a success exit flag.\nHowever, the two-norm of the final residual is higher than tolerance ({norm_residual:0.2e}>{self.tol:0.2e})"

        # Calculate elapsed time
        self.elapsed_time = time.perf_counter() - start_time

        # Print report footer
        self._print_convergence_progress(self.x_final)
        self._write_footer()

        return self.x_final

    def residual(self, x, called_from_jac=False):
        """
        Evaluate the nonlinear system residuals.

        Parameters
        ----------
        x : array_like
            Independent variable vector.

        Returns
        -------
        array_like
            Residuals of the problem.
        """

        # Increase total counter (includes finite differences)
        self.func_count_tot += 1

        # Compute problem residuals
        self.residuals_last = self.problem.residual(x)

        # Update progress report
        if not called_from_jac:
            self.func_count += 1  # Does not include finite differences
            if self.update_on == "function":
                self._print_convergence_progress(x)

        return self.residuals_last

    def gradient(self, x):
        """
        Evaluates the Jacobian of the nonlinear system of equations at the specified point x.

        This method will use the `gradient` method of the NonlinearSystemProblem class if it exists.
        If the `gradient` method is not implemented the Jacobian is appoximated using forward finite differences.

        Parameters
        ----------
        x : array-like
            Vector of independent variables.

        Returns
        -------
        array-like
            Jacobian matrix of the residual vector formatted as a 2D array.
        """

        # Evaluate the Jacobian of the residual vector
        self.grad_count += 1
        if hasattr(self.problem, "gradient"):
            # If the problem has its own Jacobian method, use it
            jacobian = self.problem.gradient(x)

        else:
            # Fall back to finite differences
            fun = lambda x: self.residual(x, called_from_jac=True)
            jacobian = numerical_differentiation.approx_gradient(
                fun,
                x,
                f0=self.residuals_last,
                method=self.derivative_method,
                abs_step=self.derivative_abs_step,
            )

        # Update progress report
        if self.update_on == "gradient":
            self._print_convergence_progress(x)

        return np.atleast_1d(jacobian)

    def _write_header(self):
        """
        Print a formatted header for the root finding report.

        This internal method is used to display a consistent header format at the
        beginning of the root finding process. The header includes columns for residual evaluations,
        gradient evaluations, norm of the residuals, and norm of the steps.
        """
        # Define header text
        initial_message = (
            f" Solve system of equations for {type(self.problem).__name__}\n"
            f" Root-finding algorithm employed: {self.method}"
        )
        self.header = f" {'Func-eval':>15}{'Grad-eval':>15}{'Norm of residual':>24}{'Norm of step':>24} "
        separator = "-" * len(self.header)
        lines_to_output = [
            separator,
            initial_message,
            separator,
            self.header,
            separator,
        ]

        # Display to stdout
        if self.display:
            for line in lines_to_output:
                print(line)

        # Write to log
        if self.logger:
            for line in lines_to_output:
                self.logger.info(line)

        # Store text in memory
        self.solution_report.extend(lines_to_output)

    def _print_convergence_progress(self, x):
        """
        Print the current solution status and update convergence history.

        This method captures and prints the following metrics:
        - Number of function evaluations
        - Number of gradient evaluations
        - Two-norm of residual vector
        - Two-norm of solution step

        The method also updates the stored convergence history for potential future analysis.

        Parameters
        ----------
        x : array-like
            The current solution (i.e., vector of independent variable values)

        Notes
        -----
        The norm of the update step is calculated as the two-norm of the difference
        between the current and the last independent variables. The norm of the residual
        vector is computed using the two-norm.
        """
        # Compute norm of residual vector
        residual = self.residuals_last
        norm_residual = np.linalg.norm(residual)
        # norm_residual = np.max([np.linalg.norm(residual), np.finfo(float).eps])

        # Compute the norm of the last step
        norm_step = np.linalg.norm(x - self.x_last) if self.x_last is not None else 0
        self.x_last = x.copy()

        # Store convergence status
        self.convergence_history["grad_count"].append(self.grad_count)
        self.convergence_history["func_count"].append(self.func_count)
        self.convergence_history["func_count_total"].append(self.func_count_tot)
        self.convergence_history["norm_residual"].append(norm_residual)
        self.convergence_history["norm_step"].append(norm_step)

        # Current convergence message
        status = f" {self.func_count:15d}{self.grad_count:15d}{norm_residual:24.6e}{norm_step:24.6e} "

        # Display to stdout
        if self.display:
            print(status)

        # Write to log
        if self.logger:
            self.logger.info(status)

        # Store text in memory
        self.solution_report.append(status)

        # Refresh the plot with current values
        if self.plot:
            self._plot_callback()

        # Evaluate callback function
        if self.callback_func:
            self.callback_func_call_count += 1
            self.callback_func(x, self.callback_func_call_count)

    def _write_footer(self):
        """
        Print a formatted footer for the root finding report.

        This method displays the final root finding result, including the
        exit message, success status, and decision variables.

        Notes
        -----
        The footer's structure is intended to match the header's style,
        providing a consistent look to the root finding report.
        """
        # Define footer text
        separator = "-" * len(self.header)
        exit_message = f"Exit message: {self.message}"
        success_status = f"Success: {self.success}"
        time_message = f"Solution time: {self.elapsed_time:.3f} seconds"
        solution_header = "Solution:"
        solution_vars = [f"   x{i} = {x:+6e}" for i, x in enumerate(self.x_final)]
        lines_to_output = [separator, success_status, exit_message, time_message]
        if self.include_solution_in_footer:
            lines_to_output += [solution_header]
            lines_to_output += solution_vars
        lines_to_output += [separator, ""]

        # Display to stdout
        if self.display:
            for line in lines_to_output:
                print(line)

        # Write to log
        if self.logger:
            for line in lines_to_output:
                self.logger.info(line)

        # Store text in memory
        self.solution_report.extend(lines_to_output)

    def _plot_callback(self, initialize=False):
        """
        Callback function to dynamically update the convergence progress plot.

        This method initializes a matplotlib plot on the first iteration and updates
        the data for each subsequent iteration. The plot showcases the evolution of
        the residual vector as a function of number of iterations.

        Note:
            This is an internal method, meant to be called within the optimization process.
        """

        # Initialize figure before first iteration
        if initialize:
            self.fig, self.ax = plt.subplots()
            (self.obj_line,) = self.ax.plot([], [], color="#D95319", marker="o")
            self.ax.set_xlabel("Number of iterations")
            self.ax.set_ylabel("Two-norm of residual vector")
            self.ax.set_yscale(self.plot_scale)
            # self.ax.xaxis.set_major_locator(MultipleLocator(1))
            # self.ax.legend(loc="upper right")
            self.fig.tight_layout(pad=1)

        # Update plot data with current values
        iteration = (
            self.convergence_history["func_count"]
            if self.update_on == "function"
            else self.convergence_history["grad_count"]
        )
        residual = self.convergence_history["norm_residual"]
        self.obj_line.set_xdata(iteration)
        self.obj_line.set_ydata(residual)

        # Adjust the plot limits
        self.ax.relim()
        self.ax.autoscale_view()

        # Redraw the plot
        plt.draw()
        plt.pause(0.01)  # small pause to allow for update

    def print_convergence_history(
        self, savefile=False, filename=None, output_dir="output"
    ):
        """
        Print the convergence history of the problem.

        The convergence history includes:
            - Number of function evaluations
            - Number of gradient evaluations
            - Two-norm of the residual vector
            - Two-norm of the solution step

        The method provides a detailed report on:
            - Exit message
            - Success status
            - Execution time

        This method should be called only after the optimization problem has been solved, as it relies on data generated by the solving process.

        Parameters
        ----------
        savefile : bool, optional
            If True, the convergence history will be saved to a file, otherwise printed to standard output. Default is False.
        filename : str, optional
            The name of the file to save the convergence history. If not specified, the filename is automatically generated
            using the problem name and the start datetime. The file extension is not required.
        output_dir : str, optional
            The directory where the plot file will be saved if savefile is True. Default is "output".

        Raises
        ------
        ValueError
            If this method is called before the problem has been solved.

        """
        if self.x_final is not None:
            if savefile:
                # Create output directory if it does not exist
                if not os.path.exists(output_dir):
                    os.makedirs(output_dir)

                # Give a name to the file if it is not specified
                if filename is None:
                    filename = f"convergence_{type(self.problem).__name__}_{self.start_datetime}.txt"

                # Write report to file
                fullfile = os.path.join(output_dir, filename)
                with open(fullfile, "w") as f:
                    f.write("\n".join(self.solution_report))

            else:
                for line in self.solution_report:
                    print(line)

        else:
            raise ValueError(
                "This method can only be used after invoking the 'solve()' method."
            )

    def plot_convergence_history(
        self, savefile=False, filename=None, output_dir="output"
    ):
        """
        Plot the convergence history of the problem as the two-norm of the residual vector versus the number of iterations.

        This method should be called only after the optimization problem has been solved, as it relies on data generated by the solving process.

        Parameters
        ----------
        savefile : bool, optional
            If True, the plot is saved to a file instead of being displayed. Default is False.
        filename : str, optional
            The name of the file to save the plot to. If not specified, the filename is automatically generated
            using the problem name and the start datetime. The file extension is not required.
        output_dir : str, optional
            The directory where the plot file will be saved if savefile is True. Default is "output".

        Returns
        -------
        matplotlib.figure.Figure
            The Matplotlib figure object for the plot. This can be used for further customization or display.

        Raises
        ------
        ValueError
            If this method is called before the problem has been solved.
        """
        if self.x_final is not None:
            self._plot_callback(initialize=True)
        else:
            raise ValueError(
                "This method can only be used after invoking the 'solve()' method."
            )

        if savefile:
            # Create output directory if it does not exist
            if not os.path.exists(output_dir):
                os.makedirs(output_dir)

            # Give a name to the file if it is not specified
            if filename is None:
                filename = (
                    f"convergence_{type(self.problem).__name__}_{self.start_datetime}"
                )

            # Save plots
            fullfile = os.path.join(output_dir, filename)
            savefig_in_formats(self.fig, fullfile, formats=[".png", ".svg"])

        return self.fig


class NonlinearSystemProblem(ABC):
    """
    Abstract base class for root-finding problems.

    Derived root-finding problem objects must implement the following method:

    - `residual`: Evaluate the system of equations for a given set of decision variables.

    Additionally, specific problem classes can define the `gradient` method to compute the Jacobians. If this method is not present in the derived class, the solver will revert to using forward finite differences for Jacobian calculations.

    Methods
    -------
    residual(x)
        Evaluate the system of equations for a given set of decision variables.

    Examples
    --------
    Here's an example of how to derive from `RootFindingProblem`::

        class MyRootFindingProblem(RootFindingProblem):
            def residual(self, x):
                # Implement evaluation logic here
                pass
    """

    @abstractmethod
    def residual(self, x):
        """
        Evaluate the system of equations for given decision variables.

        Parameters
        ----------
        x : array-like
            Vector of decision variables.

        Returns
        -------
        array_like
            Vector containing the values of the system of equations for the given decision variables.
        """
        pass



# Highlight exception messages
# https://stackoverflow.com/questions/25109105/how-to-colorize-the-output-of-python-errors-in-the-gnome-terminal/52797444#52797444
try:
    import IPython.core.ultratb
except ImportError:
    # No IPython. Use default exception printing.
    pass
else:
    import sys
    sys.excepthook = IPython.core.ultratb.ColorTB()

# Import subpackages
from .fluent_automation import *
from .fluid_properties import *
from .barotropic_model import *

# Import submodules
from .math import *
from .graphics import *
from .sCO2_utilities import *
from .utilities import *

# Package info
__version__ = "0.0.3"
PACKAGE_NAME = "barotropy"
URL_GITHUB = "https://github.com/turbo-sim/barotropy"
URL_DOCS = "https://turbo-sim.github.io/barotropy/"
URL_DTU = "https://thermalpower.dtu.dk/"
BREAKLINE = 80 * "-"

def print_banner():
    """Prints a banner."""
    banner = """
                ____                   __                        
               / __ )____ __________  / /__________  ____  __  __
              / __  / __ `/ ___/ __ \/ __/ ___/ __ \/ __ \/ / / /
             / /_/ / /_/ / /  / /_/ / /_/ /  / /_/ / /_/ / /_/ / 
            /_____/\__,_/_/   \____/\__/_/   \____/ .___/\__, /  
                                                 /_/    /____/   
"""
#     banner = """
#             ____  ___    ____  ____  __________  ____  ______  __
#            / __ )/   |  / __ \/ __ \/_  __/ __ \/ __ \/ __ \ \/ /
#           / __  / /| | / /_/ / / / / / / / /_/ / / / / /_/ /\  / 
#          / /_/ / ___ |/ _, _/ /_/ / / / / _, _/ /_/ / ____/ / /  
#         /_____/_/  |_/_/ |_|\____/ /_/ /_/ |_|\____/_/     /_/   
                                                                
# """
    print(BREAKLINE)
    print(banner)
    print(BREAKLINE)
    # https://manytools.org/hacker-tools/ascii-banner/
    # Style: Sland


def print_package_info():
    """Prints package information with predefined values."""

    info = f""" Version:       {__version__}
 Repository:    {URL_GITHUB}
 Documentation: {URL_DOCS}"""
    print_banner()
    print(info)
    print(BREAKLINE)








import pandas as pd
import numpy as np

from .fluid_properties import fluid_interface as props

from typing import Dict, Tuple


PROPERTY_LIST = ["p", "T", "rho", "gamma", "Z", "a", "h", "s"]


def compute_inlet_state(row, fluid):

    # Check for existence of columns and non-NaN values, then calculate accordingly
    if "pressure_in" in row and "temperature_in" in row and not pd.isna(row["pressure_in"]) and not pd.isna(row["temperature_in"]):
        p_in = row["pressure_in"] * 1e5  # bar to Pa
        T_in = row["temperature_in"] + 273.15  # degC to K
        fluid.set_state(props.PT_INPUTS, p_in, T_in)
    elif "pressure_in" in row and "density_in" in row and not pd.isna(row["pressure_in"]) and not pd.isna(row["density_in"]):
        d_in = row["density_in"]
        p_in = row["pressure_in"] * 1e5  # bar to Pa
        fluid.set_state(props.DmassP_INPUTS, d_in, p_in)
    elif "density_in" in row and "temperature_in" in row and not pd.isna(row["density_in"]) and not pd.isna(row["temperature_in"]):
        d_in = row["density_in"]
        T_in = row["temperature_in"] + 273.15  # degC to K
        fluid.set_state(props.DmassT_INPUTS, d_in, T_in)
    elif "pressure_in" in row and "entropy_in" in row and not pd.isna(row["pressure_in"]) and not pd.isna(row["entropy_in"]):
        s_in = row["entropy_in"] * 1e3  # kJ/kg/K to J/kg/K
        p_in = row["pressure_in"] * 1e5  # bar to Pa
        fluid.set_state(props.PSmass_INPUTS, p_in, s_in)
    elif "temperature_in" in row and "entropy_in" in row and not pd.isna(row["temperature_in"]) and not pd.isna(row["entropy_in"]):
        s_in = row["entropy_in"] * 1e3  # kJ/kg/K to J/kg/K
        T_in = row["temperature_in"] + 273.15  # degC to K
        fluid.set_state(props.SmassT_INPUTS, s_in, T_in)
    else:
        raise ValueError("Inlet state must be specified as 'pressure-temperature', 'pressure-density', or 'temperature-density' and the respective columns must exist.")

    # Append '_in' to property names and return
    inlet_props = {
        key + "_in": value
        for key, value in fluid.properties.items()
        if key in PROPERTY_LIST
    }

    return pd.Series(inlet_props)


def compute_outlet_state_isentropic(row, fluid):
    # Calculate inlet properties with p-T function call
    p_in = row["pressure_in"] * 1e5  # bar to Pa
    T_in = row["temperature_in"] + 273.15  # degC to K
    state_in = fluid.set_state(props.PT_INPUTS, p_in, T_in)

    # Calculate isentropic exit state
    p_out = row["pressure_ratio"] * p_in
    state_out = fluid.set_state(props.PSmass_INPUTS, p_out, state_in.s)

    # Append '_out' to exit properties and include dh_is
    outlet_props = {
        key + "_out": value
        for key, value in state_out.to_dict().items()
        if key in PROPERTY_LIST
    }
    outlet_props["isentropic_work"] = (state_out.h - state_in.h)/1e3

    return pd.Series(outlet_props)


def compute_critical_state(row, fluid):
    # Calculate inlet properties with p-T function call
    p_in = row["pressure_in"] * 1e5  # bar to Pa
    T_in = row["temperature_in"] + 273.15  # degC to K

    # Calculate critical state (i.e., sonic state for an isentropic expansion)
    state_critical = fluid.compute_sonic_state(props.PT_INPUTS, p_in, T_in)

    # Append '_out' to property names
    critical_props = {
        key + "_crit": value
        for key, value in state_critical.to_dict().items()
        if key in PROPERTY_LIST
    }

    return pd.Series(critical_props)


def calculate_scaling_variables(
    props: Dict[str, float]
) -> Tuple[float, float, float, float, float]:
    p = props["p"]
    T = props["T"]
    Z = props["Z"]  # Setting 1 here yields bad matching for HeRo compressor
    g = props["gamma"]
    v = np.sqrt(2 * g / (g + 1) * Z * T)
    f = g * (2 / (g + 1)) ** (g / (g - 1.0))
    a = props["a"]
    rho = props["rho"]
    return p, v, f, a, rho


def compute_equivalent_from_actual(
    row: pd.Series,
    state_ref: Dict[str, float],
    method: str = "perfect_gas",
) -> pd.Series:
    """Compute equivalent variables from actual conditions and reference state"""

    # Common calculations
    state_in = {col.rstrip("_in"): row[col] for col in row.index if col.endswith("_in")}
    p_1, v_1, f_1, a_1, rho_1 = calculate_scaling_variables(state_in)
    p_2, v_2, f_2, a_2, rho_2 = calculate_scaling_variables(state_ref)

    # Compute equivalent variables based on the selected method
    variables = {}
    if method == "perfect_gas":
        delta = p_1 / p_2
        theta = (v_1 / v_2) ** 2
        epsilon = f_2 / f_1
        variables["delta"] = delta
        variables["theta"] = theta
        variables["epsilon"] = epsilon
        variables["mass_flow_eq"] = row["mass_flow"] * (np.sqrt(theta) * epsilon / delta)
        variables["isentropic_work_eq"] = row["isentropic_work"] * (1 / theta)
        variables["angular_speed_eq"] = row["angular_speed"] * (1 / np.sqrt(theta))
    elif method == "real_gas":
        variables["mass_flow_eq"] = row["mass_flow"] * (rho_2 * a_2) / (rho_1 * a_1)
        variables["isentropic_work_eq"] = row["isentropic_work"] * (a_2 / a_1) ** 2
        variables["angular_speed_eq"] = row["angular_speed"] * (a_2 / a_1)

    return pd.Series(variables)


def compute_actual_from_equivalent(
    row: pd.Series,
    state_ref: Dict[str, float],
    method: str = "perfect_gas",
) -> pd.Series:
    """Compute equivalent variables from actual conditions and reference state"""

    # Common calculations
    state_in = {col.rstrip("_in"): row[col] for col in row.index if col.endswith("_in")}
    p_1, v_1, f_1, a_1, rho_1 = calculate_scaling_variables(state_in)
    p_2, v_2, f_2, a_2, rho_2 = calculate_scaling_variables(state_ref)

    # Compute equivalent variables based on the selected method
    variables = {}
    if method == "perfect_gas":
        delta = p_1 / p_2
        theta = (v_1 / v_2) ** 2
        epsilon = f_2 / f_1
        variables["delta"] = delta
        variables["theta"] = theta
        variables["epsilon"] = epsilon
        variables["mass_flow"] = row["mass_flow_eq"] * (np.sqrt(theta) * epsilon / delta) ** -1
        # variables["isentropic_work"] = row["isentropic_work_eq"] * (1 / theta) ** -1
        # variables["angular_speed"] = row["angular_speed_eq"] * (1 / np.sqrt(theta)) ** -1
    elif method == "real_gas":
        variables["mass_flow"] = row["mass_flow_eq"] * ((rho_2 * a_2) / (rho_1 * a_1))  ** -1
        # variables["isentropic_work"] = row["isentropic_work_eq"] * (a_2 / a_1) ** -2
        # variables["angular_speed"] = row["angular_speed_eq"] * (a_2 / a_1) ** -1

    return pd.Series(variables)


import matplotlib.pyplot as plt
import matplotlib.colors as mcolors

def plot_ps_diagram(fluid, ax, range_x, range_y, range_z):

    # Define nice colormap
    colormap = mcolors.LinearSegmentedColormap.from_list(
        "truncated_blues", plt.cm.Blues(np.linspace(0.4, 1.0, 256))
    )

    # Define limits
    prop_x = "s"
    prop_y = "p"
    prop_z = "T"
    prop_pair = props.PSmass_INPUTS

    # Plot phase diagram
    ax = fluid.plot_phase_diagram(
        prop_x,
        prop_y,
        plot_critical_point=True,
        plot_quality_isolines=True,
        plot_pseudocritical_line=True,
    )

    # Plot pressure isobars
    prop_dict = props.compute_property_grid(fluid, prop_pair, range_y, range_x)
    contour = ax.contour(
        prop_dict[prop_x],
        prop_dict[prop_y],
        prop_dict[prop_z],
        range_z,
        cmap=colormap,
        linewidths=0.5,
    )

    # Colorbar settings
    sm = plt.cm.ScalarMappable(
        cmap=colormap,
        norm=plt.Normalize(vmin=min(range_z), vmax=max(range_z)),
    )
    sm.set_array([])  # This line is necessary for ScalarMappable
    cbar = plt.colorbar(sm, ax=ax)
    cbar.ax.set_ylabel("Temperature [$^\circ$C]", rotation=90, labelpad=20)
    cbar.set_ticks(range_z)  # Set tick positions
    cbar.set_ticklabels([f"{level-273.15:.0f}" for level in range_z])

    return ax


def plot_Ts_diagram(fluid, ax, range_x, range_y, range_z):

    # Define nice colormap
    colormap = mcolors.LinearSegmentedColormap.from_list(
        "truncated_blues", plt.cm.Blues(np.linspace(0.4, 1.0, 256))
    )

    # Define limits
    prop_x = "s"
    prop_y = "T"
    prop_z = "p"
    prop_pair = props.SmassT_INPUTS
    
    # Plot phase diagram
    ax = fluid.plot_phase_diagram(
        prop_x,
        prop_y,
        plot_critical_point=True,
        plot_quality_isolines=True,
        plot_pseudocritical_line=True,
    )

    # Plot pressure isobars
    prop_dict = props.compute_property_grid(fluid, prop_pair, range_x, range_y)
    contour = ax.contour(
        prop_dict[prop_x],
        prop_dict[prop_y],
        prop_dict[prop_z],
        range_z,
        cmap=colormap,
        linewidths=0.5,
    )

    # Colorbar settings
    sm = plt.cm.ScalarMappable(
        cmap=colormap,
        norm=plt.Normalize(vmin=min(range_z), vmax=max(range_z)),
    )
    sm.set_array([])  # This line is necessary for ScalarMappable
    cbar = plt.colorbar(sm, ax=ax)
    cbar.ax.set_ylabel("Pressure [bar]", rotation=90, labelpad=20)
    cbar.set_ticks(range_z)  # Set tick positions
    cbar.set_ticklabels([f"{level/1e5:.0f}" for level in range_z])

    return ax



import os
import logging
import numpy as np
import matplotlib as mpl
import matplotlib.pyplot as plt

from cycler import cycler
from datetime import datetime

COLORS_PYTHON = [
    "#1f77b4",
    "#ff7f0e",
    "#2ca02c",
    "#d62728",
    "#9467bd",
    "#8c564b",
    "#e377c2",
    "#7f7f7f",
    "#bcbd22",
    "#17becf",
]

COLORS_MATLAB = [
    "#0072BD",
    "#D95319",
    "#EDB120",
    "#7E2F8E",
    "#77AC30",
    "#4DBEEE",
    "#A2142F",
]


def set_plot_options(
    fontsize=13,
    grid=False,
    major_ticks=True,
    minor_ticks=True,
    margin=0.05,
    color_order="matlab",
    linewidth=1.25,
):
    """
    Set options for creating publication-quality figures using Matplotlib.

    This function updates the internal Matplotlib settings to better align with standards for publication-quality figures.
    Features include improved font selections, tick marks, grid appearance, and color selections.

    Parameters
    ----------
    fontsize : int, optional
        Font size for text elements in the plot. Default is 13.
    grid : bool, optional
        Whether to show grid lines on the plot. Default is True.
    major_ticks : bool, optional
        Whether to show major ticks. Default is True.
    minor_ticks : bool, optional
        Whether to show minor ticks. Default is True.
    margin : float, optional
        Margin size for axes. Default is 0.05.
    color_order : str, optional
        Color order to be used for plot lines. Options include "python" and "matlab". Default is "matlab".

    """

    if isinstance(color_order, str):
        if color_order.lower() == "python":
            color_order = [
                "#1f77b4",
                "#ff7f0e",
                "#2ca02c",
                "#d62728",
                "#9467bd",
                "#8c564b",
                "#e377c2",
                "#7f7f7f",
                "#bcbd22",
                "#17becf",
            ]

        elif color_order.lower() == "matlab":
            color_order = [
                "#0072BD",
                "#D95319",
                "#EDB120",
                "#7E2F8E",
                "#77AC30",
                "#4DBEEE",
                "#A2142F",
            ]

    # Define dictionary of custom settings
    rcParams = {
        "text.usetex": False,
        "font.size": fontsize,
        "font.style": "normal",
        "font.family": "serif",  # 'serif', 'sans-serif', 'cursive', 'fantasy', 'monospace'
        "font.serif": ["Times New Roman"],  # ['times new roman', 'cmr10']
        "mathtext.fontset": "stix",  # ["stix", 'cm']
        "axes.edgecolor": "black",
        "axes.linewidth": 1.25,
        "axes.titlesize": fontsize,
        "axes.titleweight": "normal",
        "axes.titlepad": fontsize * 1.4,
        "axes.labelsize": fontsize,
        "axes.labelweight": "normal",
        "axes.labelpad": fontsize,
        "axes.xmargin": margin,
        "axes.ymargin": margin,
        # "axes.zmargin": margin,
        "axes.grid": grid,
        "axes.grid.axis": "both",
        "axes.grid.which": "major",
        "axes.prop_cycle": cycler(color=color_order),
        "grid.alpha": 0.5,
        "grid.color": "black",
        "grid.linestyle": "-",
        "grid.linewidth": 0.5,
        "legend.borderaxespad": 1,
        "legend.borderpad": 0.6,
        "legend.edgecolor": "black",
        "legend.facecolor": "white",
        "legend.labelcolor": "black",
        "legend.labelspacing": 0.3,
        "legend.fancybox": True,
        "legend.fontsize": fontsize - 2,
        "legend.framealpha": 1.00,
        "legend.handleheight": 0.7,
        "legend.handlelength": 1.25,
        "legend.handletextpad": 0.8,
        "legend.markerscale": 1.0,
        "legend.numpoints": 1,
        "lines.linewidth": linewidth,
        "lines.markersize": 4,
        "lines.markeredgewidth": 1.25,
        "lines.markerfacecolor": "white",
        "xtick.direction": "in",
        "xtick.labelsize": fontsize - 1,
        "xtick.bottom": major_ticks,
        "xtick.top": major_ticks,
        "xtick.major.size": 6,
        "xtick.major.width": 1.25,
        "xtick.minor.size": 3,
        "xtick.minor.width": 0.75,
        "xtick.minor.visible": minor_ticks,
        "ytick.direction": "in",
        "ytick.labelsize": fontsize - 1,
        "ytick.left": major_ticks,
        "ytick.right": major_ticks,
        "ytick.major.size": 6,
        "ytick.major.width": 1.25,
        "ytick.minor.size": 3,
        "ytick.minor.width": 0.75,
        "ytick.minor.visible": minor_ticks,
        "savefig.dpi": 500,
    }

    # Update the internal Matplotlib settings dictionary
    mpl.rcParams.update(rcParams)


def print_installed_fonts():
    """
    Print the list of fonts installed on the system.

    This function identifies and prints all available fonts for use in Matplotlib.
    """
    fonts = mpl.font_manager.findSystemFonts(fontpaths=None, fontext="ttf")
    for font in sorted(fonts):
        print(font)


def print_rc_parameters(filename=None):
    """
    Print the current rcParams used by Matplotlib or write to file if provided.

    This function provides a quick overview of the active configuration parameters within Matplotlib.
    """
    params = mpl.rcParams
    for key, value in params.items():
        print(f"{key}: {value}")

    if filename:
        with open(filename, "w") as file:
            for key, value in params.items():
                file.write(f"{key}: {value}\n")


def savefig_in_formats(fig, path_without_extension, formats=[".png", ".svg"], dpi=500):
    """
    Save a given Matplotlib figure in multiple file formats.

    Parameters
    ----------
    fig : matplotlib.figure.Figure
        The figure object to be saved.
    path_without_extension : str
        The full path to save the figure excluding the file extension.
    formats : list of str, optional
        A list of string file extensions to specify which formats the figure should be saved in.
        Supported formats are ['.png', '.svg', '.pdf', '.eps'].
        Default is ['.png', '.svg'].
    dpi : int, optional
        The resolution in dots per inch with which to save the image. This parameter affects only PNG files.
        Default is 500.

    Examples
    --------
    >>> import matplotlib.pyplot as plt
    >>> fig, ax = plt.subplots()
    >>> ax.plot([0, 1], [0, 1])
    >>> save_fig_in_formats(fig, "/path/to/figure/filename")

    This will save the figure as "filename.png", "filename.svg", and "filename.pdf" in the "/path/to/figure/" directory.
    """

    # Validate export formats
    allowed_formats = {".png", ".svg", ".pdf", ".eps"}
    invalid_formats = set(formats) - allowed_formats
    if invalid_formats:
        raise ValueError(
            f"Unsupported file formats: {', '.join(invalid_formats)}. Allowed formats: {', '.join(allowed_formats)}"
        )

    for ext in formats:
        if ext == ".png":
            # Save PNG files with specified DPI
            fig.savefig(f"{path_without_extension}{ext}", bbox_inches="tight", dpi=dpi)
        else:
            # Save in vector formats without DPI setting
            fig.savefig(f"{path_without_extension}{ext}", bbox_inches="tight")


def create_logger(name, path=None, use_datetime=True):
    """
    Creates and configures a logging object for recording logs during program execution.

    Parameters
    ----------
    name : str
        Name of the log file. Allows for differentiation when analyzing logs from different components or runs of a program.
    path : str, optional
        Specifies the directory where the log files will be saved. By default, a directory named "logs"
        will be created in the current working directory (cwd).
    use_datetime : bool, optional
        Determines whether the log filename should have a unique datetime identifier appended. Default is True.

    Returns
    -------
    logger : object
        Configured logger object.

    Notes
    -----
    - By default, the function sets the log level to `INFO`, which means the logger will handle
      messages of level `INFO` and above (like `ERROR`, `WARNING`, etc.). The log entries will contain
      the timestamp, log level, and the actual log message.
    - When `use_datetime=True`, each log file will have a unique datetime identifier. This ensures
      traceability, avoids overwriting previous logs, allows chronological ordering of log files, and
      handles concurrency in multi-instance environments.
    """

    # Define logs directory if it is not provided
    if path is None:
        path = os.path.join(os.getcwd(), "logs")

    # Create logs directory if it does not exist
    if not os.path.exists(path):
        os.makedirs(path)

    # Define file name and path
    if use_datetime:
        current_time = datetime.now().strftime("%Y-%m-%d_%H-%M-%S")
        log_filename = os.path.join(path, f"{name}_{current_time}.log")
    else:
        log_filename = os.path.join(path, f"{name}.log")

    # Set logger configuration
    logging.basicConfig(
        filename=log_filename,
        level=logging.INFO,
        format="%(asctime)s - %(levelname)s - %(message)s",
    )

    # Create logger object
    logger = logging.getLogger()

    return logger



import os
import numpy as np
from scipy.optimize import minimize

# Attempt to import pygmo and pygmo_plugins_nonfree
try:
    import pygmo as pg
    import pygmo_plugins_nonfree as ppnf

    PYGMO_AVAILABLE = True

except ImportError:
    PYGMO_AVAILABLE = False

SCIPY_SOLVERS = [
    "nelder-mead",
    "powell",
    "cg",
    "bfgs",
    "newton-cg",
    "l-bfgs-b",
    "tnc",
    "cobyla",
    "slsqp",
    "trust-constr",
    "dogleg",
    "trust-ncg",
    "trust-krylov",
    "trust-exact",
]
"""List of valid Scipy solvers"""


PYGMO_SOLVERS = [
    "ipopt",
    "snopt",
    # "worhp",
]
"""List of valid Pygmo solvers"""


NLOPT_SOLVERS = [
    "auglag",
    "auglag_eq",
    "bobyqa",
    "ccsaq",
    "cobyla",
    "lbfgs",
    "mma",
    "neldermead",
    "newuoa",
    "newuoa_bound",
    "praxis",
    "sbplx",
    "slsqp",
    "tnewton",
    "tnewton_precond",
    "tnewton_precond_restart",
    "tnewton_restart",
    "var1",
    "var2",
]
"""List of valid NLOpt solvers (interfaced through Pygmo)"""


def minimize_scipy(problem, x0, method, options):
    """
    Optimize a given problem using the specified SciPy method.

    Parameters
    ----------
    problem : OptimizationProblem
        An object representing the problem.
    x0 : array_like
        Initial guess for the decision variables.
    method : str
        The name of the Scipy solver to use. Must be one of the solvers defined in the SCIPY_SOLVERS list.
    options : dict
        Configuration options for the solver. Options vary depending on the solver used.

    Returns
    -------
    tuple
        A tuple (success, message) where 'success' is a boolean indicating if the optimization was successful, and 'message' is a string describing the outcome.

    Raises
    ------
    ValueError
        If an invalid solver is specified.
    """

    # Check if method is valid
    if method not in SCIPY_SOLVERS:
        error_message = (
            f"Invalid solver: '{method}'. \nAvailable options:\n   - "
            + "\n   - ".join(SCIPY_SOLVERS)
            + "."
        )
        raise ValueError(error_message)

    # Define the options dictionary
    options = options.copy()  # Work with a copy to avoid side effects
    tol = options.pop("tol")
    max_iter = options.pop("max_iter")
    default_options = {
        "bfgs": {"gtol": tol, "maxiter": max_iter},
        "l-bfgs-b": {"ftol": tol, "gtol": tol, "maxiter": max_iter},
        "slsqp": {"ftol": tol, "maxiter": max_iter},
        "trust-constr": {"gtol": tol, "xtol": tol, "maxiter": max_iter},
    }
    default_options = (
        default_options[method] if method in default_options.keys() else {}
    )
    combined_options = default_options | options  # Merge defaults with input

    # Convert bounds from Pygmo to Scipy convention
    if problem.get_bounds() is None:
        bounds = None
    else:
        bounds = convert_pygmo_to_scipy_bounds(problem.get_bounds)()

    # Define list of constraint dictionaries
    constr = []
    if problem.get_nec() > 0:
        constr.append({"type": "eq", "fun": problem.c_eq, "jac": problem.c_eq_jac})
    if problem.get_nic() > 0:
        # Convert inequality constraint convention from Pygmo (c<0) to SciPy (c>0)
        c_ineq = lambda x: -problem.c_ineq(x)
        c_ineq_jac = lambda x: -problem.c_ineq_jac(x)
        constr.append({"type": "ineq", "fun": c_ineq, "jac": c_ineq_jac})

    # Solve the optimization problem
    solution = minimize(
        problem.f,
        x0,
        tol=tol,
        jac=problem.f_jac,
        constraints=constr,
        bounds=bounds,
        method=method,
        options=combined_options,
    )

    # Optimization output
    success = solution.success
    message = solution.message

    return success, message


def minimize_pygmo(problem, x0, method, options):
    """
    Optimize a given problem using the specified Pygmo optimization method.

    Parameters
    ----------
    problem : OptimizationProblem
        An object representing the problem.
    x0 : array_like
        Initial guess for the decision variables.
    method : str
        The name of the Pygmo solver to use. Must be one of the solvers defined in the PYGMO_SOLVERS list.
    options : dict
        Configuration options for the solver. Options vary depending on the solver used.

    Returns
    -------
    tuple
        A tuple (success, message) where 'success' is a boolean indicating if the optimization was successful, and 'message' is a string describing the outcome.

    Raises
    ------
    ValueError
        If an invalid solver name is specified.
    """

    # Check if pygmo is available
    if not PYGMO_AVAILABLE:
        raise ImportError(
            f"To use the selected optimizer ({method}), you need to install 'pygmo' via Conda.\n\n"
            "   1. Activate the Conda virtual environment where 'turboflow' is installed. For example:\n\n"
            "           conda activate turboflow_env\n\n"
            "   2. Install 'pygmo' and 'pygmo_plugins_nonfree' by running the following command:\n\n"
            "           conda install --channel conda-forge pygmo pygmo_plugins_nonfree\n"
        )

    if method == "snopt":
        return _minimize_pygmo_snopt(problem, x0, options)
    elif method == "ipopt":
        return _minimize_pygmo_ipopt(problem, x0, options)
    # elif method == "worhp":
    #     return _minimize_pygmo_worhp(problem, x0, options)
    else:
        error_message = (
            f"Invalid solver: '{method}'. \nAvailable options:\n   - "
            + "\n   - ".join(PYGMO_SOLVERS)
            + "."
        )
        raise ValueError(error_message)


def _minimize_pygmo_ipopt(problem, x0, options):
    """Solve optimization problem using Pygmo's wrapper to IPOPT"""

    # Define mapping from exit flag to status
    exitflag_mapping = {
        0: "Solve Succeeded",
        1: "Solved To Acceptable Level",
        2: "Infeasible Problem Detected",
        3: "Search Direction Becomes Too Small",
        4: "Diverging Iterates",
        5: "User Requested Stop",
        6: "Feasible Point Found",
        -1: "Maximum Iterations Exceeded",
        -2: "Restoration Failed",
        -3: "Error In Step Computation",
        -4: "Maximum CpuTime Exceeded",
        -5: "Maximum WallTime Exceeded",
        -10: "Not Enough Degrees Of Freedom",
        -11: "Invalid Problem Definition",
        -12: "Invalid Option",
        -13: "Invalid Number Detected",
        -100: "Unrecoverable Exception",
        -101: "NonIpopt Exception Thrown",
        -102: "Insufficient Memory",
        -199: "Internal Error",
    }

    # Define solver options
    options = options.copy()  # Work with a copy to avoid side effects
    tol = options.pop("tol")
    max_iter = options.pop("max_iter")
    if tol < 1e-3:
        print(
            f"IPOPT struggles to converge to tight tolerances when exact Hessians are not provided.\nConsider relaxing the termination tolerance above 1e-3. Current tolerance: {tol:0.2e}."
        )

    # Define solver options
    default_options = {
        "print_level": 0,  # No output to the screen
        "sb": "yes",  # Suppress banner
        "hessian_approximation": "limited-memory",  # exact, limited-memory
        "limited_memory_update_type": "bfgs",  # bfgs, sr1
        "line_search_method": "cg-penalty",  # filter, cg-penalty, penalty
        "limited_memory_max_history": 30,  # bfgs
        "max_iter": int(max_iter),
        "tol": tol,
    }
    combined_options = default_options | options

    # Solve the problem
    problem = pg.problem(problem)
    algorithm = pg.algorithm(pg.ipopt())
    algorithm_handle = algorithm.extract(pg.ipopt)
    _set_pygmo_options(algorithm_handle, combined_options)
    population = pg.population(problem, size=1)
    population.set_x(0, x0)
    population = algorithm.evolve(population)

    # Optimization output
    exitflag = algorithm_handle.get_last_opt_result()
    message = exitflag_mapping.get(exitflag, "Unknown Status")
    success = exitflag in (0, 1)

    return success, message


def _minimize_pygmo_snopt(problem, x0, options):
    """Solve optimization problem using Pygmo's wrapper to SNOPT"""

    # Define solver options
    options = options.copy()  # Work with a copy to avoid side effects
    tol = options.pop("tol")
    max_iter = options.pop("max_iter")
    default_options = {
        "Major feasibility tolerance": tol,
        "Major optimality tolerance": tol,
        "Minor feasibility tolerance": tol,
        "Iterations limit": int(max_iter),
        "Major iterations limit": 1e6,
        "Minor iterations limit": 1e6,
        "Hessian updates": 30,
    }
    combined_options = default_options | options

    # Define SNOPT path
    lib = os.getenv("SNOPT_LIB")
    if not lib:
        raise ValueError("SNOPT library path not set in environment variables.")

    # Solve the problem
    problem = pg.problem(problem)
    algorithm = pg.algorithm(ppnf.snopt7(library=lib, minor_version=7))
    algorithm_handle = algorithm.extract(ppnf.snopt7)
    _set_pygmo_options(algorithm_handle, combined_options)
    population = pg.population(problem, size=1)
    population.set_x(0, x0)
    population = algorithm.evolve(population)

    # Optimization output

    message = _extract_snopt_message(algorithm.get_extra_info())
    success = (
        True
        if message == "Finished successfully - optimality conditions satisfied"
        else False
    )

    return success, message


def _minimize_pygmo_worhp(problem, x0, options):
    """Solve optimization problem using Pygmo's wrapper to WORHP"""

    # Define solver options
    options = options.copy()  # Work with a copy to avoid side effects
    tol = options.pop("tol")
    max_iter = options.pop("max_iter")
    default_options = {
        "Algorithm": 1,  # Set: 1 (SQP) or 2 (IP)
        "BFGSmethod": 0,  # Set 0 to use the classic dense BFGS method
        "TolOpti": tol,
        "TolFeas": tol,
        "MaxIter": max_iter,
    }
    combined_options = default_options | options

    # Define WORHP path
    lib = os.getenv("WORHP_LIB")
    if not lib:
        raise ValueError("WORHP library path not set in environment variables.")

    # Solve the problem
    problem = pg.problem(problem)
    algorithm = pg.algorithm(ppnf.worhp(screen_output=False, library=lib))
    algorithm_handle = algorithm.extract(ppnf.worhp)
    _set_pygmo_options(algorithm_handle, combined_options)
    population = pg.population(problem, size=1)
    population.set_x(0, x0)
    population = algorithm.evolve(population)

    # Optimization output
    success = ""
    message = ""

    return success, message


def minimize_nlopt(problem, x0, method, options):
    """
    Optimize a given problem using the specified NLOpt optimization method.

    Parameters
    ----------
    problem : OptimizationProblem
        An object representing the problem.
    x0 : array_like
        Initial guess for the decision variables.
    method : str
        The name of the Pygmo solver to use. Must be one of the solvers defined in the NLOPT_SOLVERS list.
    options : dict
        Configuration options for the solver. Options vary depending on the solver used.

    Returns
    -------
    tuple
        A tuple (success, message) where 'success' is a boolean indicating if the optimization was successful, and 'message' is a string describing the outcome.

    Raises
    ------
    ValueError
        If an invalid solver name is specified.
    """

    # Check if method is valid
    if method not in NLOPT_SOLVERS:
        error_message = (
            f"Invalid solver: '{method}'. \nAvailable options:\n   - "
            + "\n   - ".join(NLOPT_SOLVERS)
            + "."
        )
        raise ValueError(error_message)

    # Define exit flag mapping
    exitflag_mapping = {
        1: "Generic success return value.",
        2: "Optimization stopped because stop value was reached.",
        3: "Optimization stopped because relative or absolute tolerance on function value was reached.",
        4: "Optimization stopped because relative or absolute tolerance on x was reached.",
        5: "Optimization stopped because maximum evaluations limit was reached.",
        6: "Optimization stopped because maximum time limit was reached.",
        -1: "Generic failure code.",
        -2: "Invalid arguments (e.g., lower bounds are bigger than upper bounds, unknown algorithm specified, etc.).",
        -3: "Ran out of memory.",
        -4: "Halted due to roundoff errors limiting progress. Optimization might still return a useful result.",
        -5: "Optimization was forcibly stopped.",
    }

    # Define solver options
    options = options.copy()  # Work with a copy to avoid side effects
    tol = options.pop("tol")
    max_iter = options.pop("max_iter")
    algorithm = pg.algorithm(pg.nlopt(solver=method))
    handle = algorithm.extract(pg.nlopt)
    handle.maxeval = max_iter
    handle.ftol_rel = tol
    handle.xtol_rel = tol

    # Solve the problem
    problem = pg.problem(problem)
    population = pg.population(problem, size=1)
    population.set_x(0, x0)
    population = algorithm.evolve(population)

    # Optimization output
    exitflag = handle.get_last_opt_result()
    message = exitflag_mapping.get(exitflag, "Unknown Status")
    success = exitflag in (1, 2, 3, 4)

    return success, message


def convert_pygmo_to_scipy_bounds(pygmo_bounds):
    """
    Create a function that converts bounds from Pygmo format to SciPy format.

    Parameters
    ----------
    pygmo_method : callable
        A method that, when called, returns a tuple containing two lists: the first list of lower bounds and the second of upper bounds (Pygmo convention).

    Returns
    -------
    callable
        A function that, when called, returns a list of tuples where each tuple contains the lower and upper bound for each decision variable (Scipy convention).

    """

    def converter():
        lower_bounds, upper_bounds = pygmo_bounds()
        scipy_bounds = list(zip(lower_bounds, upper_bounds))
        return scipy_bounds

    return converter


def _extract_snopt_message(info_str):
    # Search for the start of the relevant message
    start = info_str.find("Last optimisation return code: ")
    if start == -1:
        return "Optimization return code not found."

    # Adjust the start position to the actual start of the message
    start += len("Last optimisation return code: ")

    # Find the end of the line
    end = info_str.find("\n", start)
    if end == -1:
        end = len(info_str)

    # Extract and return the message
    return info_str[start:end].strip()


def _set_pygmo_options(handle, options):
    """
    Sets multiple options on a PyGMO solver handle based on the types of values provided in a dictionary.

    This function supports various PyGMO solvers including IPOPT, WORHP, and SNOPT. It dynamically calls
    the appropriate method to set each option (string, integer, or numeric) based on the type of each value
    in the `options` dictionary.

    Parameters
    ----------
    handle : object
        The PyGMO solver handle. This object should have methods for setting options,
        such as `set_string_option`, `set_integer_option`, and `set_numeric_option`.
    options : dict
        A dictionary where each key is the option name and each value is the option value.
        The function determines the method to use based on the type of the value:
        `str` for `set_string_option`, `int` for `set_integer_option`, and `float` for `set_numeric_option`.

    Raises
    ------
    ValueError
        If the value type is not supported (not an int, float, or str).

    Examples
    --------
    >>> algorithm = pg.algorithm(pg.ipopt())
    >>> handle = algorithm.extract(pg.ipopt)
    >>> options = {
    >>>     "sb": "yes",
    >>>     "print_level": 0,
    >>>     "tol": 1e-6,
    >>>     "max_iter": 1000
    >>> }
    >>> set_pygmo_options(handle, options)
    """
    for option_name, value in options.items():
        if isinstance(value, str):
            handle.set_string_option(option_name, value)
        elif isinstance(value, int):
            handle.set_integer_option(option_name, value)
        elif isinstance(value, float):
            handle.set_numeric_option(option_name, value)
        else:
            raise ValueError(
                "Unsupported option type: {} for option {}".format(
                    type(value), option_name
                )
            )



import sys
import time

def _print_transcript_real_time(filename, frequency=0.1):
    """Stream new lines from a file in real time."""
    with open(filename, 'r') as file:
        # Move the pointer to the end of the file
        file.seek(0, 2)
        while True:
            line = file.readline()
            if line:
                print('\t', line, end='')
            else:
                time.sleep(frequency)

if __name__ == "__main__":

    # Get the transcript file as command-line argument
    if len(sys.argv) != 2:
        print("Usage: python stream_output.py <file_path>")
        sys.exit(1)

    # Display transcript as the file gets updated  
    _print_transcript_real_time(filename=sys.argv[1])
